"""Logging tests."""

import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives


@parametrize_directives
@pytest.mark.parametrize('missing_argument', ('start', 'end'))
def test_start_end_arguments_not_found(
    directive,
    missing_argument,
    page,
    tmp_path,
    plugin,
    caplog,
):
    included_file_name = 'included.md'
    includer_file_name = 'includer.md'
    included_file = tmp_path / included_file_name
    includer_file = tmp_path / includer_file_name

    includer_content = f'''# Heading

{{%
  {directive} "{included_file}"
  start="<!--start-->"
  end="<!--end-->"
%}}
'''
    if missing_argument == 'end':
        included_content = '<!--start-->Included content'
    else:
        included_content = 'Included content<!--end-->'

    includer_file.write_text(includer_content)
    included_file.write_text(included_content)

    expected_result = '''# Heading

Included content
'''

    assert on_page_markdown(
        includer_content, page(includer_file), tmp_path, plugin,
    ) == expected_result

    assert (
        f"Delimiter {missing_argument} '<!--{missing_argument}-->' of"
        f" '{directive}' directive at {includer_file_name}:3"
        f' not detected in the file {included_file_name}'
    ) in caplog.text



import os

import pytest
from mkdocs.exceptions import PluginError

import mkdocs_include_markdown_plugin.cache
from mkdocs_include_markdown_plugin import IncludeMarkdownPlugin
from mkdocs_include_markdown_plugin.cache import (
    Cache,
    get_cache_directory,
    initialize_cache,
    is_platformdirs_installed,
)
from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import FakeConfig, parametrize_directives


@pytest.mark.parametrize(
    ('url', 'expected_result'),
    (
        (
            'https://raw.githubusercontent.com/mondeja/mkdocs-include-markdown-plugin/master/examples/basic/mkdocs.yml',  # noqa: E501
            '''site_name: Foo
plugins:
  - include-markdown
''',
        ),
        (
            'https://raw.githubusercontent.com/mondeja/mkdocs-include-markdown-plugin/master/examples/basic/docs/included.md',  # noqa: E501
            '''Some ignored content.

<--start-->

Some included content.
''',
        ),
    ),
)
@parametrize_directives
def test_page_included_by_url_is_cached(
    directive,
    url,
    expected_result,
    page,
    tmp_path,
    plugin,
):
    if not is_platformdirs_installed():
        assert initialize_cache(600, '') is None
        return

    cache_dir = get_cache_directory('')
    os.makedirs(cache_dir, exist_ok=True)

    file_path = os.path.join(
        cache_dir, Cache.generate_unique_key_from_url(url),
    )
    if os.path.isfile(file_path):
        os.remove(file_path)

    cache = Cache(cache_dir, 600)

    def run():
        return on_page_markdown(
            f'''{{%
    {directive} "{url}"
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
            http_cache=cache,
        )

    assert run() == expected_result

    assert os.path.isfile(file_path)
    run()
    assert os.path.isfile(file_path)

    os.remove(file_path)


def test_cache_setting_when_not_available_raises_error(monkeypatch):
    monkeypatch.setattr(
        mkdocs_include_markdown_plugin.cache,
        'is_platformdirs_installed',
        lambda: False,
    )
    plugin = IncludeMarkdownPlugin()
    plugin.config = FakeConfig(cache=600, cache_dir='')
    with pytest.raises(PluginError) as exc:
        plugin.on_config({})
    assert (
        'Either `cache_dir` global setting must be configured or'
        ' `platformdirs` package is required'
    ) in str(exc.value)


def test_cache_setting_available_with_cache_dir(monkeypatch):
    monkeypatch.setattr(
        mkdocs_include_markdown_plugin.cache,
        'is_platformdirs_installed',
        lambda: False,
    )
    plugin = IncludeMarkdownPlugin()
    plugin.config = FakeConfig(cache=600, cache_dir='foo')
    plugin.on_config({})



"""Plugin configuration."""

from __future__ import annotations

from mkdocs.config.base import Config
from mkdocs.config.config_options import (
    ListOfItems,
    Optional,
    Type as MkType,
)


class PluginConfig(Config):  # noqa: D101
    opening_tag = MkType(str, default='{%')
    closing_tag = MkType(str, default='%}')
    encoding = MkType(str, default='utf-8')
    preserve_includer_indent = MkType(bool, default=True)
    dedent = MkType(bool, default=False)
    trailing_newlines = MkType(bool, default=True)
    comments = MkType(bool, default=False)
    rewrite_relative_urls = MkType(bool, default=True)
    heading_offset = MkType(int, default=0)
    order = MkType(str, default='alpha-path')
    start = Optional(MkType(str))
    end = Optional(MkType(str))
    exclude = ListOfItems(MkType(str), default=[])
    cache = MkType(int, default=0)
    cache_dir = MkType(str, default='')
    recursive = MkType(bool, default=True)
    directives = MkType(
        dict,
        default={
            '__default': '',
            'include': 'include',
            'include-markdown': 'include-markdown',
        },
    )



"""Nested inclusion tests."""

import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown


@pytest.mark.parametrize(
    (
        'first_includer_content',
        'second_includer_content',
        'included_content',
        'expected_result',
        'expected_warnings_schemas',
    ),
    (
        # Includer -> Markdown -> Markdown
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
%}''',
            '''Some text from second includer.

{%
  include-markdown "{filepath}"
%}
''',
            '''Some test from final included.''',
            '''# Header

Some text from second includer.

Some test from final included.
''',
            [],
            id='includer -> markdown -> markdown',
        ),

        # Includer -> Markdown -> file
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
%}''',
            '''Some text from second includer.

{%
  include "{filepath}"
%}
''',
            '''Some test from final included.''',
            '''# Header

Some text from second includer.

Some test from final included.
''',
            [],
            id='includer -> markdown -> file',
        ),

        # Includer -> file -> file
        pytest.param(
            '''# Header

{%
  include "{filepath}"
%}''',
            '''Some text from second includer.

{%
  include "{filepath}"
%}
''',
            '''Some test from final included.''',
            '''# Header

Some text from second includer.

Some test from final included.
''',
            [],
            id='includer -> file -> file',
        ),

        # Includer -> file -> Markdown
        pytest.param(
            '''# Header

{%
  include "{filepath}"
%}''',
            '''Some text from second includer.

{%
  include-markdown "{filepath}"
  start="<!-- start -->"
  end="<!-- end -->"
%}
''',
            '''This must be ignored
<!-- start -->Some test from final included.<!-- end -->

This must be ignored also
''',
            '''# Header

Some text from second includer.

Some test from final included.
''',
            [],
            id='includer -> file -> markdown',
        ),

        # cumulative_heading_offset
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=1
%}''',
            '''# Header 2

{%
  include-markdown "{filepath}"
  heading-offset=1
%}
''',
            '''# Header 3
''',
            '''# Header

## Header 2

### Header 3

''',
            [],
            id='cumulative_heading_offset',
        ),

        # start and end defined in first inclusion but not found
        pytest.param(
            '''# Header

{%
  include-markdown '{filepath}'
  start="<!--start-->"
  end="<!--end-->"
%}''',
            '''# Header 2

''',
            '''# Header 3
''',
            '''# Header

# Header 2

''',
            [
                (
                    "Delimiter start '<!--start-->' of 'include-markdown'"
                    ' directive at {first_includer_file}:3 not detected'
                    ' in the file {second_includer_file}'
                ),
                (
                    "Delimiter end '<!--end-->' of 'include-markdown'"
                    ' directive at {first_includer_file}:3 not detected'
                    ' in the file {second_includer_file}'
                ),
            ],
            id='start-end-not-found (first-level)',
        ),
        # start and end defined in second inclusion but not found
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
%}''',
            '''# Header 2

{%
  include-markdown "{filepath}"
  start="<!--start-->"
  end="<!--end-->"
%}''',
            '''# Header 3

Included content
''',
            '''# Header

# Header 2

# Header 3

Included content
''',
            [
                (
                    "Delimiter start '<!--start-->' of 'include-markdown'"
                    ' directive at {second_includer_file}:3 not detected'
                    ' in the file {included_file}'
                ),
                (
                    "Delimiter end '<!--end-->' of 'include-markdown'"
                    ' directive at {second_includer_file}:3 not detected'
                    ' in the file {included_file}'
                ),
            ],
            id='start-end-not-found (second-level)',
        ),
        # recursive inclusion disabled with `include` directive
        pytest.param(
            '''# Header

{%
  include "{filepath}"
  recursive=false
%}''',
            '''# Header 2

{% include "{filepath}" %}
''',
            '''# Header 3

This content must not be included.
''',
            '''# Header

# Header 2

{% include "{filepath}" %}
''',
            [],
            id='include-recursive=false',
        ),
        # recursive inclusion disabled with `include-markdown` directive
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  recursive=false
%}''',
            '''# Header 2

{% include-markdown "{filepath}" %}
''',
            '''# Header 3

This content must not be included.
''',
            '''# Header

# Header 2

{% include-markdown "{filepath}" %}
''',
            [],
            id='include-markdown-recursive=false',
        ),
    ),
)
def test_nested_include(
    first_includer_content,
    second_includer_content,
    included_content,
    expected_result,
    expected_warnings_schemas,
    page,
    plugin,
    caplog,
    tmp_path,
):
    first_includer_file = tmp_path / 'first-includer.txt'
    second_includer_file = tmp_path / 'second-includer.txt'
    included_file = tmp_path / 'included.txt'

    first_includer_content = first_includer_content.replace(
        '{filepath}', second_includer_file.as_posix(),
    )
    second_includer_content = second_includer_content.replace(
        '{filepath}', included_file.as_posix(),
    )
    expected_result = expected_result.replace(
        '{filepath}', included_file.as_posix(),
    )

    first_includer_file.write_text(first_includer_content)
    second_includer_file.write_text(second_includer_content)
    included_file.write_text(included_content)

    # assert content
    assert on_page_markdown(
        first_includer_content, page(first_includer_file), tmp_path, plugin,
    ) == expected_result

    # assert warnings
    expected_warnings_schemas = expected_warnings_schemas or []
    expected_warnings = [
        msg_schema.replace(
            '{first_includer_file}',
            str(first_includer_file.relative_to(tmp_path)),
        ).replace(
            '{second_includer_file}',
            str(second_includer_file.relative_to(tmp_path)),
        ).replace(
            '{included_file}',
            str(included_file.relative_to(tmp_path)),
        ) for msg_schema in expected_warnings_schemas
    ]

    for record in caplog.records:
        assert record.msg in expected_warnings
    assert len(expected_warnings_schemas) == len(caplog.records)


def test_nested_include_relpath(page, tmp_path, plugin):
    docs_dir = tmp_path / 'docs'
    docs_dir.mkdir()

    first_includer_file = tmp_path / 'first-includer.txt'
    second_includer_file = docs_dir / 'second-includer.txt'
    included_file = tmp_path / 'included.txt'

    first_includer_content = '''# Header

{%
  include-markdown "./docs/second-includer.txt"
%}
'''
    first_includer_file.write_text(first_includer_content)

    second_includer_content = '''Text from second includer.

{%
  include-markdown "../included.txt"
%}
'''
    second_includer_file.write_text(second_includer_content)

    included_file.write_text('Included content.')

    expected_result = '''# Header

Text from second includer.

Included content.

'''

    assert on_page_markdown(
        first_includer_content,
        page(first_includer_file),
        docs_dir,
        plugin,
    ) == expected_result



import os
import sys

import pytest


TESTS_DIR = os.path.abspath(os.path.dirname(__file__))
SRC_DIR = os.path.abspath(os.path.join(TESTS_DIR, '..', 'src'))
for d in (SRC_DIR, TESTS_DIR):
    if d not in sys.path:
        sys.path.insert(0, d)

from mkdocs_include_markdown_plugin import IncludeMarkdownPlugin  # noqa: E402


@pytest.fixture
def page():
    """Fake mkdocs page object."""
    def _page(file_path):
        return type(
            'FakeMkdocsPage', (), {
                'file': type(
                    'FakeMdocsPageFile', (), {
                        'abs_src_path': file_path,
                    },
                ),
            },
        )
    return _page


@pytest.fixture
def plugin(request):
    """Populate a plugin, with optional indirect config parameter."""
    plugin = IncludeMarkdownPlugin()
    errors, warnings = plugin.load_config(getattr(request, 'param', {}))
    assert errors == []
    assert warnings == []
    return plugin



"""String processing tests."""

import pytest

from mkdocs_include_markdown_plugin.cache import Cache
from mkdocs_include_markdown_plugin.process import (
    increase_headings_offset,
    read_url,
    rewrite_relative_urls,
)


@pytest.mark.parametrize(
    ('markdown', 'source_path', 'destination_path', 'expected_result'),
    (
        # Markdown Relative Links
        pytest.param(
            "Here's a [link](CHANGELOG.md) to the changelog.",
            'README',
            'docs/nav.md',
            "Here's a [link](../CHANGELOG.md) to the changelog.",
            id='relative-link',
        ),
        pytest.param(
            '''Here's a [link whose text is really long and so is broken across
multiple lines](CHANGELOG.md) to the changelog.
''',
            'README',
            'docs/nav.md',
            '''Here's a [link whose text is really long and so is broken across
multiple lines](../CHANGELOG.md) to the changelog.
''',
            id='multiline-link',
        ),
        pytest.param(
            '''
Check [this link](foobar.md) for more information
''',
            'docs/includes/feature_a/index.md',
            'docs/setup.md',
            '''
Check [this link](includes/feature_a/foobar.md) for more information
''',
            id='relative-link-down',
        ),
        pytest.param(
            '''Here's a [link](CHANGELOG.md#v1.2.3) to the changelog.
''',
            'README',
            'docs/nav.md',
            '''Here's a [link](../CHANGELOG.md#v1.2.3) to the changelog.
''',
            id='link-with-hash',
        ),
        pytest.param(
            '''Here's a [link][changelog] to the changelog.

[changelog]: CHANGELOG.md
''',
            'README',
            'docs/nav.md',
            '''Here's a [link][changelog] to the changelog.

[changelog]: ../CHANGELOG.md
''',
            id='link-reference',
        ),
        pytest.param(
            "Here's a diagram: ![diagram](assets/diagram.png)",
            'README',
            'docs/home.md',
            "Here's a diagram: ![diagram](../assets/diagram.png)",
            id='image',
        ),
        pytest.param(
            'Build status: [![Build Status](badge.png)](build/)',
            'README',
            'docs/home.md',
            'Build status: [![Build Status](../badge.png)](../build/)',
            id='image-inside-link',
        ),
        # HTML Relative Links
        pytest.param(
            ('Here\'s a diagram: <img id="foo" src="assets/diagram.png"'
             ' alt="diagram" class="bar" />'),
            'README',
            'docs/home.md',
            ('Here\'s a diagram: <img id="foo" src="../assets/diagram.png"'
             ' alt="diagram" class="bar" />'),
            id='html-image',
        ),
        pytest.param(
            ('Here\'s a diagram: <source id="foo" src="assets/diagram.png"'
             ' class="bar" />'),
            'README',
            'docs/home.md',
            ('Here\'s a diagram: <source id="foo" src="../assets/diagram.png"'
             ' class="bar" />'),
            id='html-source',
        ),
        pytest.param(
            ('Here\'s a diagram: <a id="foo" href="badge.png"'
             ' class="bar">example</a>'),
            'README',
            'docs/home.md',
            ('Here\'s a diagram: <a id="foo" href="../badge.png"'
             ' class="bar">example</a>'),
            id='html-anchor',
        ),
        pytest.param(
            ("Here's a diagram: <img id='foo' src='assets/diagram.png'"
             " alt='diagram' class='bar' />"),
            'README',
            'docs/home.md',
            ("Here's a diagram: <img id='foo' src='../assets/diagram.png'"
             " alt='diagram' class='bar' />"),
            id='html-image-single-quote',
        ),
        pytest.param(
            ("Here's a diagram: <a id='foo' href='assets/diagram.png'"
             " class='bar'>example</a>"),
            'README',
            'docs/home.md',
            ("Here's a diagram: <a id='foo' href='../assets/diagram.png'"
             " class='bar'>example</a>"),
            id='html-anchor-single-quote',
        ),
        # HTML Relative Links Adversarial tests:
        # (attribute contains >, attribute without value, multiple tag in line)
        pytest.param(
            ('<img id="foo" attr="3>2" attr2 src="assets/diagram.png"'
             ' alt="diagram" class="bar" /><img id="foo" attr="3>2"'
             ' src="assets/diagram.png" alt="diagram" class="bar" />'),
            'README',
            'docs/home.md',
            ('<img id="foo" attr="3>2" attr2 src="../assets/diagram.png"'
             ' alt="diagram" class="bar" /><img id="foo" attr="3>2"'
             ' src="../assets/diagram.png" alt="diagram" class="bar" />'),
            id='html-image-adversarial-test',
        ),
        pytest.param(
            ('<a id="foo" attr="3>2" attr2 href="badge.png" class="bar">'
             'foo</a><a id="foo" attr="3>2" href="badge.png" class="bar">'
             'bar</a>'),
            'README',
            'docs/home.md',
            ('<a id="foo" attr="3>2" attr2 href="../badge.png" class="bar">'
             'foo</a><a id="foo" attr="3>2" href="../badge.png" class="bar">'
             'bar</a>'),
            id='html-anchor-adversarial-test',
        ),
        # HTML Relative Links Adversarial test: img no end slash
        pytest.param(
            ('Here\'s a diagram: <img id="foo" src="assets/diagram.png"'
             ' alt="diagram" class="bar">'),
            'README',
            'docs/home.md',
            ('Here\'s a diagram: <img id="foo" src="../assets/diagram.png"'
             ' alt="diagram" class="bar">'),
            id='html-image-no-end-slash',
        ),
        # Non-relative links
        pytest.param(
            "Here's a [link](/CHANGELOG.md) to the changelog.",
            'README',
            'docs/nav.md',
            "Here's a [link](/CHANGELOG.md) to the changelog.",
            id='absolute-link',
        ),
        pytest.param(
            'A [link](https://example.com/index.html) to the changelog.',
            'README',
            'docs/nav.md',
            'A [link](https://example.com/index.html) to the changelog.',
            id='external-link',
        ),
        pytest.param(
            "Here's a [link](https://example.com) to the changelog.",
            'README',
            'docs/nav.md',
            "Here's a [link](https://example.com) to the changelog.",
            id='external-top-level-link',
        ),
        pytest.param(
            ('<img id="foo" attr="3>2" src="https://example.com/image.png"'
             ' class="bar" />'),
            'README',
            'docs/home.md',
            ('<img id="foo" attr="3>2" src="https://example.com/image.png"'
             ' class="bar" />'),
            id='html-image-external-link',
        ),
        pytest.param(
            ('<a id="foo" attr="3>2" href="https://example.com/index.html"'
             ' class="bar" />'),
            'README',
            'docs/home.md',
            ('<a id="foo" attr="3>2" href="https://example.com/index.html"'
             ' class="bar" />'),
            id='html-anchor-external-link',
        ),
        pytest.param(
            '<a id="ab" attr="3>2" href="https://example.com" class="cd" />',
            'README',
            'docs/home.md',
            '<a id="ab" attr="3>2" href="https://example.com" class="cd" />',
            id='html-anchor-external-top-level-link',
        ),
        pytest.param(
            '''[Homepage](/) [Github](https://github.com/user/repo)
[Privacy policy](/privacy)''',
            'README',
            'docs/nav.md',
            '''[Homepage](/) [Github](https://github.com/user/repo)
[Privacy policy](/privacy)''',
            id='absolute-urls',
        ),
        pytest.param(
            '[contact us](mailto:hello@example.com)',
            'README',
            'docs/nav.md',
            '[contact us](mailto:hello@example.com)',
            id='mailto-urls',
        ),
        pytest.param(
            '''Some text before

```cpp
// Some code in which rewrites shouldn't be proccessed.
// https://github.com/mondeja/mkdocs-include-markdown-plugin/issues/78
const auto lambda = []() { .... };
```
''',
            'README',
            'examples/lambda.md',
            '''Some text before

```cpp
// Some code in which rewrites shouldn't be proccessed.
// https://github.com/mondeja/mkdocs-include-markdown-plugin/issues/78
const auto lambda = []() { .... };
```
''',
            id='cpp-likelink-fenced-codeblock',
        ),
        pytest.param(
            (
                'Text before\n'
                '    \n    '
                'const auto lambda = []() { .... };\n    \nText after\n'
            ),
            'README',
            'examples/lambda.md',
            (
                'Text before\n'
                '    \n    '
                'const auto lambda = []() { .... };\n    \nText after\n'
            ),
            id='cpp-likelink-indented-codeblock',
        ),
        pytest.param(
            (
                'Text before\r\n'
                '    \r\n    '
                'const auto lambda = []() { .... };\r\n    \r\nText after\r\n'
            ),
            'README',
            'examples/lambda.md',
            (
                'Text before\r\n'
                '    \r\n    '
                'const auto lambda = []() { .... };\r\n    \r\nText after\r\n'
            ),
            id='cpp-likelink-indented-codeblock-windows-newlines',
        ),
        pytest.param(
            '''```
[link](CHANGELOG.md)
```
''',
            'README',
            'docs/nav.md',
            '''```
[link](CHANGELOG.md)
```
''',
            id='exclude-fenced-code-blocks',
        ),
        pytest.param(
            '''```
<img src="assets/diagram.png" alt="diagram">
<a href="badge.png">example</a>
```
''',
            'README',
            'docs/nav.md',
            '''```
<img src="assets/diagram.png" alt="diagram">
<a href="badge.png">example</a>
```
''',
            id='exclude-fenced-code-blocks-html',
        ),
        pytest.param(
            (
                '    \n'
                '    [link](CHANGELOG.md)\n'
                '    \n'
            ),
            'README',
            'docs/nav.md',
            (
                '    \n'
                '    [link](CHANGELOG.md)\n'
                '    \n'
            ),
            id='exclude-indented-code-blocks',
        ),
        pytest.param(
            (
                '    \n'
                '    [link](CHANGELOG.md)\n'
            ),
            'README',
            'docs/nav.md',
            # is rewritten because not newline at end of code block
            (
                '    \n'
                '    [link](../CHANGELOG.md)\n'
            ),
            id='exclude-indented-code-blocks-eof',
        ),
        pytest.param(
            (
                '    [link](CHANGELOG.md)\n'
                '    \n'
            ),
            'README',
            'docs/nav.md',
            (
                '    [link](../CHANGELOG.md)\n'
                '    \n'
            ),
            # No newline before, is not an indented code block, see:
            # https://spec.commonmark.org/0.28/#indented-code-blocks
            id='no-exclude-indented-code-blocks-missing-newline-before',
        ),
        pytest.param(
            (
                '    \n'
                '    [link](CHANGELOG.md)\n'
                'Foo\n'
            ),
            'README',
            'docs/nav.md',
            (
                '    \n'
                '    [link](../CHANGELOG.md)\n'
                'Foo\n'
            ),
            # No newline after, is not an indented code block, see:
            # https://spec.commonmark.org/0.28/#indented-code-blocks
            id='no-exclude-indented-code-blocks-missing-newline-after',
        ),
        # Internal anchor in included file
        # https://github.com/mondeja/mkdocs-include-markdown-plugin/issues/266
        pytest.param(
            'This is a link to an [internal anchor](#internal-anchor).',
            'README',
            'docs/nav.md',
            'This is a link to an [internal anchor](#internal-anchor).',
            id='internal-anchor',
        ),
    ),
)
def test_rewrite_relative_urls(
    markdown,
    source_path,
    destination_path,
    expected_result,
):
    assert rewrite_relative_urls(
        markdown,
        source_path,
        destination_path,
    ) == expected_result


@pytest.mark.parametrize(
    ('markdown', 'offset', 'expected_result'),
    (
        pytest.param(
            '''# Foo

```python
# this is a comment
hello = "world"
```

    # this is an indented
    codeblock

- This list item has a fenced codeblock inside:

   ```
   # fenced codeblock inside list item
   ```

# Bar

Some text

## Baz
''',
            2,
            '''### Foo

```python
# this is a comment
hello = "world"
```

    # this is an indented
    codeblock

- This list item has a fenced codeblock inside:

   ```
   # fenced codeblock inside list item
   ```

### Bar

Some text

#### Baz
''',
            id='```',
        ),
        pytest.param(
            '''# Foo

~~~python
# this is a comment
hello = "world"
~~~

# Bar

Some text

## Baz
''',
            3,
            '''#### Foo

~~~python
# this is a comment
hello = "world"
~~~

#### Bar

Some text

##### Baz
''',
            id='~~~',
        ),
        pytest.param(
            '''# Foo

~~~python
# this is a comment
hello = "world"
~~~

# Bar

Some text

## Baz

```
# another comment
```

# Qux
''',
            1,
            '''## Foo

~~~python
# this is a comment
hello = "world"
~~~

## Bar

Some text

### Baz

```
# another comment
```

## Qux
''',
            id='```,~~~',
        ),
        pytest.param(
            '''# Foo

    # this is a comment
    hello = "world"

# Bar

    # another comment

\t# comment in tabbed indented codeblock\r\n
## Qux
''',
            1,
            '''## Foo

    # this is a comment
    hello = "world"

## Bar

    # another comment

\t# comment in tabbed indented codeblock\r\n
### Qux
''',
            id='indented-codeblocks',
        ),
    ),
)
def test_dont_increase_heading_offset_inside_fenced_codeblocks(
    markdown,
    offset,
    expected_result,
):
    assert increase_headings_offset(markdown, offset=offset) == expected_result


def test_read_url_cached_content(tmp_path):
    url = (
        'https://raw.githubusercontent.com/mondeja/'
        'mkdocs-include-markdown-plugin/master/README.md'
    )
    cache_dir = tmp_path.as_posix()
    cached_file_name = Cache.generate_unique_key_from_url(url)
    cached_file_path = tmp_path / cached_file_name
    if cached_file_path.exists():
        cached_file_path.unlink()

    cache = Cache(cache_dir, 600)
    content = read_url(url, cache)
    assert cached_file_path.exists()

    cached_content = cached_file_path.read_text(
        encoding='utf-8',
    ).split('\n', 1)[1]
    assert content == cached_content

    assert cache.get_(url) == cached_content
    assert cache.get_(url) == read_url(url, cache)
    cached_file_path.unlink()



"""``include-markdown`` directive tests"""

import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown
from mkdocs_include_markdown_plugin.placeholders import build_placeholder


@pytest.mark.parametrize(
    (
        'includer_schema',
        'content_to_include',
        'expected_result_schema',
        'expected_warnings_schemas',
    ),
    (
        # Simple case
        pytest.param(
            '# Header\n\n{% include-markdown "{filepath}" %}\n',
            'This must be included.',
            '''# Header

This must be included.
''',
            [],
            id='simple case',
        ),

        # Start option
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  start="<!--start-here-->"
%}
''',
            '''This must be ignored.
<!--start-here-->
This must be included.''',
            '''# Header


This must be included.
''',
            [],
            id='start',
        ),

        # End option
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  end="<!--end-here-->"
%}
''',
            '''This must be included.
<!--end-here-->
This must be ignored.''',
            '''# Header

This must be included.

''',
            [],
            id='end',
        ),

        # Start and end options
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  start="<!--start-here-->"
  end="<!--end-here-->"
  comments=true
%}
''',
            '''This must be ignored.
<!--start-here-->
This must be included.
<!--end-here-->
This must be ignored also.''',
            '''# Header

<!-- BEGIN INCLUDE {filepath} '&lt;!--start-here--&gt;' '&lt;!--end-here--&gt;' -->

This must be included.

<!-- END INCLUDE -->
''',  # noqa: E501
            [],
            id='start/end',
        ),

        # Start and end with escaped special characters
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  start="<!--\\tstart -->"
  end="<!--\\tend -->"
  comments=true
%}
''',
            '''This must be ignored.
<!--\tstart -->
This must be included.
<!--\tend -->
This must be ignored also.''',
            '''# Header

<!-- BEGIN INCLUDE {filepath} '&lt;!--\\tstart --&gt;' '&lt;!--\\tend --&gt;' -->

This must be included.

<!-- END INCLUDE -->
''',  # noqa: E501
            [],
            id='start/end (escaped special characters)',
        ),

        # Start and end with unescaped special characters
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  start="<!--\nstart -->"
  end="<!--\nend -->"
  comments=true
%}
''',
            '''This must be ignored.
<!--\nstart -->
This must be included.
<!--\nend -->
This must be ignored also.''',
            '''# Header

<!-- BEGIN INCLUDE {filepath} '&lt;!--
start --&gt;' '&lt;!--
end --&gt;' -->

This must be included.

<!-- END INCLUDE -->
''',
            [],
            id='start/end (unescaped special characters)',
        ),

        # Exclude start and end comments
        pytest.param(
            '{% include-markdown "{filepath}" %}',
            '''Foo''',
            '''Foo''',
            [],
            id='comments=false (default)',
        ),

        # Multiples start and end matchs
        pytest.param(
            '''{%
  include-markdown "{filepath}"
  start="<!--start-tag-->"
  end="<!--end-tag-->"
%}''',
            '''Some text

<!--start-tag-->
This should be included.
<!--end-tag-->

This shouldn't be included.

<!--start-tag-->
This should be included also.
<!--end-tag-->

Here some text
that should be ignored.

<!--start-->
<!--end-->

Etc
<!--start-tag-->
This should be included even if hasn't defined after end tag.
''',
            '''
This should be included.

This should be included also.

This should be included even if hasn't defined after end tag.
''',
            [],
            id='multiple-start-end-matchs',
        ),

        # Don't preserve included indent
        pytest.param(
            '''1. Ordered list item
    {%
      include-markdown "{filepath}"
      preserve-includer-indent=false
    %}''',
            '''- Unordered sublist item
    - Other unordered sublist item''',
            '''1. Ordered list item
    - Unordered sublist item
    - Other unordered sublist item''',
            [],
            id='preserve-includer-indent=false',
        ),

        # Don't specify end and finds start in included content
        pytest.param(
            '''# Header
{%
    include-markdown "{filepath}"
    start="<!--start-->"
%}''',
            '''Some text

<!--start-->
More text
''',
            '''# Header

More text
''',
            [],
            id='start=foo-end=None',
        ),

        # Don't specify start and finds end in included content
        pytest.param(
            '''# Header
{%
    include-markdown "{filepath}"
    end='<!--end-->'
%}''',
            '''
Some text
<!--end-->
More text
''',
            '''# Header

Some text
''',
            [],
            id='start=None-end=foo',
        ),

        # Don't specify end but not finds start in included content
        pytest.param(
            '''# Header
{%
    include-markdown '{filepath}'
    start='<!--start-->'
%}''',
            '''Some text
''',
            '''# Header
''',
            [
                (
                    "Delimiter start '<!--start-->' of 'include-markdown'"
                    ' directive at {filepath}:2 not detected in the file'
                    ' {included_file}'
                ),
            ],
            id='start=foo (not found)-end=None',
        ),

        # Don't specify start but not finds end in included content
        pytest.param(
            '''# Header
{%
    include-markdown "{filepath}"
    end="<!--end-->"
%}''',
            '''
Some text
''',
            '''# Header

Some text
''',
            [
                (
                    "Delimiter end '<!--end-->' of 'include-markdown'"
                    ' directive at {filepath}:2'
                    ' not detected in the file {included_file}'
                ),
            ],
            id='start=None-end=foo (not found)',
        ),

        # Preserve includer indent
        pytest.param(
            '''1. Ordered list item
    {% include-markdown "{filepath}" %}''',
            '''- First unordered sublist item
- Second unordered sublist item
- Third unordered sublist item''',
            '''1. Ordered list item
    - First unordered sublist item
    - Second unordered sublist item
    - Third unordered sublist item''',
            [],
            id='preserve-includer-indent=true (default)',
        ),

        # Custom options ordering
        pytest.param(
            '''1. Ordered list item
    {%
      include-markdown "{filepath}"
      preserve-includer-indent=true
      end="<!--end-->"
      start="<!--start-->"
    %}''',
            '''<!--start-->- First unordered sublist item
- Second unordered sublist item<!--end-->
- Third unordered sublist item''',
            '''1. Ordered list item
    - First unordered sublist item
    - Second unordered sublist item''',
            [],
            id='custom options ordering',
        ),

        # Content unindentation
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  dedent=true
%}
''',
            '''    - Foo
    - Bar
        - Baz''',
            '''# Header

- Foo
- Bar
    - Baz
''',
            [],
            id='dedent=true',
        ),

        # Content unindentation + preserve includer indent
        pytest.param(
            '''# Header

    {%
      include-markdown "{filepath}"
      dedent=true
      preserve-includer-indent=true
    %}
''',
            '''        - Foo
        - Bar
            - Baz''',
            '''# Header

    - Foo
    - Bar
        - Baz
''',
            [],
            id='dedent=true,preserve-includer-indent=true',
        ),

        # Markdown heading offsets
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=1
%}
''',
            '''# This should be a second level heading.

Example data''',
            '''# Header

## This should be a second level heading.

Example data
''',
            [],
            id='heading-offset=1',
        ),
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=2
%}
''',
            '''# This should be a third level heading.

Example data''',
            '''# Header

### This should be a third level heading.

Example data
''',
            [],
            id='heading-offset=2',
        ),

        # Markdown heading no offset
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
%}
''',
            '''# This should be a first level heading.

Example data''',
            '''# Header

# This should be a first level heading.

Example data
''',
            [],
            id='no heading-offset (default)',
        ),

        # Markdown heading zero offset
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=0
%}
''',
            '''# This should be a first level heading.

Example data''',
            '''# Header

# This should be a first level heading.

Example data
''',
            [],
            id='heading-offset=0',
        ),

        # Markdown heading negative offset
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=-2
%}
''',
            '''#### This should be a second level heading.

Example data''',
            '''# Header

## This should be a second level heading.

Example data
''',
            [],
            id='heading-offset=-2',
        ),
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=-1
%}
''',
            '''#### This should be a third level heading.

Example data''',
            '''# Header

### This should be a third level heading.

Example data
''',
            [],
            id='heading-offset=-1',
        ),

        # Markdown heading positive offset beyond rational limits
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=90
%}
''',
            '''#### This should be a 94th level heading.

Example data''',
            '''# Header

''' + '#' * 94 + ''' This should be a 94th level heading.

Example data
''',
            [],
            id='heading-offset=90',
        ),

        # Markdown heading negative offset beyond rational limits
        pytest.param(
            '''# Header

{%
  include-markdown "{filepath}"
  heading-offset=-90
%}
''',
            '''#### This should be a first level heading.

Example data''',
            '''# Header

# This should be a first level heading.

Example data
''',
            [],
            id='heading-offset=-90',
        ),

        # Include from URL
        pytest.param(
            '''# Header

{%
  include-markdown "https://raw.githubusercontent.com/mondeja/mkdocs-include-markdown-plugin/master/examples/basic/docs/included.md"
  comments=true
%}

After include.
''',  # noqa: E501
            'foo (not used)\n',
            '''# Header

<!-- BEGIN INCLUDE https://raw.githubusercontent.com/mondeja/mkdocs-include-markdown-plugin/master/examples/basic/docs/included.md -->
Some ignored content.

<--start-->

Some included content.

<!-- END INCLUDE -->

After include.
''',  # noqa: E501
            [],
            id='url',
        ),

        # UTF-8 characters
        pytest.param(
            '''# Header

{% include-markdown "{filepath}" %}
''',
            '''Тест інклуде
азъ
два
Ббэ
[bɛ]буки
[ˈbukʲɪ]/bʲbadобаóba
Вввэ
vɛвѣди
ˈvʲedʲɪvʲвот
Гггэ
ɡɛглаголь
ɡɫɐˈɡolʲɡ/ or /gʲЖж	год god
ДдД д	дэdɐˈbro
''',
            '''# Header

Тест інклуде
азъ
два
Ббэ
[bɛ]буки
[ˈbukʲɪ]/bʲbadобаóba
Вввэ
vɛвѣди
ˈvʲedʲɪvʲвот
Гггэ
ɡɛглаголь
ɡɫɐˈɡolʲɡ/ or /gʲЖж	год god
ДдД д	дэdɐˈbro

''',
            [],
            id='russian-characters',
        ),

        # Right strip unix trailing newlines
        pytest.param(
            '''1. List item number 1
1. {% include-markdown "{filepath}" trailing-newlines=false %}
1. List item number 3
''',
            'Content to include\n',
            '''1. List item number 1
1. Content to include
1. List item number 3
''',
            [],
            id='rstrip-unix-trailing-newlines',
        ),

        # Right strip windows trailing newlines
        pytest.param(
            '''1. List item number 1
1. {%
  include-markdown "{filepath}"
  comments=false
  trailing-newlines=false
%}
1. List item number 3
''',
            'Content to include\r\n\r\n\r',
            '''1. List item number 1
1. Content to include
1. List item number 3
''',
            [],
            id='rstrip-windows-trailing-newlines',
        ),

        # Right strip trailing newlines keeping comments
        pytest.param(
            '''1. List item number 1
1. {% include-markdown "{filepath}" trailing-newlines=false comments=true %}
1. List item number 3
''',
            'Content to include\n',
            '''1. List item number 1
1. <!-- BEGIN INCLUDE {filepath} -->Content to include<!-- END INCLUDE -->
1. List item number 3
''',
            [],
            id='rstrip-trailing-newlines-comments',
        ),

        pytest.param(
            '''1. List item number 1
1. {%
  include-markdown "{filepath}"
  trailing-newlines=false
  comments=true
%}
1. List item number 3
''',
            'Content to include\n',
            '''1. List item number 1
1. <!-- BEGIN INCLUDE {filepath} -->Content to include<!-- END INCLUDE -->
1. List item number 3
''',
            [],
            id='rstrip-trailing-newlines-comments-multiline-directive',
        ),

        pytest.param(
            (
                "{% include-markdown \"{filepath}\""
                " start='<!--start-\\'including-->'"
                " comments=true"
                " %}"
            ),
            '''Ignored content

<!--start-'including-->
Content to include
''',
            r'''<!-- BEGIN INCLUDE {filepath} '&lt;!--start-&#x27;including--&gt;' '' -->

Content to include

<!-- END INCLUDE -->''',  # noqa: E501
            [],
            id='escape-comments',
        ),

        pytest.param(
            '''1.  This is the first number line

1.  {%
       include-markdown "{filepath}"
       comments=true
    %}

1.  If everything works as expected this should be number 3
''',
            '''This content chunk contains code

```
This is my example
It is a code block
```

With some text after it
''',
            '''1.  This is the first number line

1.  <!-- BEGIN INCLUDE {filepath} -->
    This content chunk contains code
''' + '    ' + '''
    ```
    This is my example
    It is a code block
    ```
''' + '    ' + '''
    With some text after it
''' + '    ' + '''
    <!-- END INCLUDE -->

1.  If everything works as expected this should be number 3
''',  # noqa: ISC003
            [],
            id='include-code-block-to-list-item (#123)',
        ),

        # Internal anchor in included file properly rewritten
        pytest.param(
            '''# Header

{% include-markdown "{filepath}" %}''',
            '''# First level heading

Example data

## Second level heading

Link to [second level heading](#second-level-heading).
''',
            '''# Header

# First level heading

Example data

## Second level heading

Link to [second level heading](#second-level-heading).
''',
            [],
            id='internal-anchor',
        ),

        # Placeholder collision
        pytest.param(
            '''# Header

''' + build_placeholder(0) + '''

{%
  include-markdown "{filepath}"
%}
''',
            'Content to include',
            '''# Header

''' + build_placeholder(0) + '''

Content to include
''',
            [],
            id='placeholder-collision',
        ),
    ),
)
def test_include_markdown(
    includer_schema,
    content_to_include,
    expected_result_schema,
    expected_warnings_schemas,
    page,
    plugin,
    caplog,
    tmp_path,
):
    included_file = tmp_path / 'included.md'
    includer_file = tmp_path / 'includer.md'

    included_file.write_text(content_to_include, encoding='utf-8')
    includer_file.write_text(
        content_to_include.replace('{filepath}', included_file.as_posix()),
        encoding='utf-8',
    )

    page_content = includer_schema.replace(
        '{filepath}', included_file.as_posix(),
    )
    includer_file.write_text(page_content, encoding='utf-8')

    # assert content
    expected_result = expected_result_schema.replace(
        '{filepath}', included_file.as_posix(),
    )

    assert on_page_markdown(
        page_content,
        page(includer_file),
        tmp_path,
        plugin,
    ) == expected_result

    # assert warnings
    expected_warnings_schemas = expected_warnings_schemas or []
    expected_warnings = [
        msg_schema.replace(
            '{filepath}',
            str(includer_file.relative_to(tmp_path)),
        ).replace(
            '{included_file}',
            str(included_file.relative_to(tmp_path)),
        ) for msg_schema in expected_warnings_schemas
    ]

    for record in caplog.records:
        assert record.msg in expected_warnings
    assert len(expected_warnings_schemas) == len(caplog.records)


@pytest.mark.parametrize('rewrite_relative_urls', ('true', 'false', None))
def test_include_markdown_relative_rewrite(
    page,
    tmp_path,
    rewrite_relative_urls,
    plugin,
):
    option_value = '' if rewrite_relative_urls is None else (
        'rewrite-relative-urls=' + rewrite_relative_urls
    )

    docs_dir = tmp_path / 'docs'
    docs_dir.mkdir()

    includer_path = tmp_path / 'includer.md'
    includer_path.write_text(
        f'''
# Heading

{{%
    include-markdown "./docs/page.md"
    start="<!--start-->"
    end="<!--end-->"
    comments=true
    {option_value}
%}}
''',
    )

    included_file_path = docs_dir / 'page.md'
    included_file_path.write_text(
        '''
# Subpage Heading
<!--start-->
Here's [a link](page2.md) and here's an image: ![](image.png)

Here's a [reference link][ref-link].

[ref-link]: page3.md
<!--end-->
''',
    )

    output = on_page_markdown(
        includer_path.read_text(),
        page(str(includer_path)),
        docs_dir,
        plugin,
    )

    if rewrite_relative_urls in ['true', None]:
        assert output == '''
# Heading

<!-- BEGIN INCLUDE ./docs/page.md '&lt;!--start--&gt;' '&lt;!--end--&gt;' -->

Here's [a link](docs/page2.md) and here's an image: ![](docs/image.png)

Here's a [reference link][ref-link].

[ref-link]: docs/page3.md

<!-- END INCLUDE -->
'''
    else:
        # include without rewriting
        assert output == '''
# Heading

<!-- BEGIN INCLUDE ./docs/page.md '&lt;!--start--&gt;' '&lt;!--end--&gt;' -->

Here's [a link](page2.md) and here's an image: ![](image.png)

Here's a [reference link][ref-link].

[ref-link]: page3.md

<!-- END INCLUDE -->
'''


def test_multiple_includes(page, tmp_path, plugin):
    snippet_filepath = tmp_path / 'snippet.md'
    another_filepath = tmp_path / 'another.md'
    includer_file = tmp_path / 'includer.md'

    includer_content = f'''# Heading

{{%
  include-markdown "{snippet_filepath}"
%}}

# Heading 2

{{%
  include-markdown "{another_filepath}"
%}}

# Heading 3

{{% include "{another_filepath}" %}}
'''
    snippet_content = 'Snippet'
    another_content = 'Another'

    includer_file.write_text(includer_content)
    snippet_filepath.write_text(snippet_content)
    another_filepath.write_text(another_content)

    expected_result = '''# Heading

Snippet

# Heading 2

Another

# Heading 3

Another
'''
    assert on_page_markdown(
        includer_content, page(includer_file), tmp_path, plugin,
    ) == expected_result



import os
import subprocess
import sys

import pytest
from mkdocs import config
from mkdocs.commands.build import build
from mkdocs.exceptions import Abort

from mkdocs_include_markdown_plugin.cache import is_platformdirs_installed
from testing_helpers import rootdir


EXAMPLES_DIR = os.path.join(rootdir, 'examples')


def config_is_using_cache_setting(config_file_path):
    with open(config_file_path, encoding='utf-8') as f:
        return 'cache:' in f.read()


@pytest.mark.parametrize('dirname', os.listdir(EXAMPLES_DIR))
def test_examples_subprocess(dirname):
    example_dir = os.path.join(EXAMPLES_DIR, dirname)
    config_file = os.path.join(example_dir, 'mkdocs.yml')
    expected_returncode = 1 if config_is_using_cache_setting(
        config_file,
    ) and not is_platformdirs_installed() else 0

    proc = subprocess.Popen(
        [sys.executable, '-mmkdocs', 'build'],
        cwd=example_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = proc.communicate()

    assert proc.returncode == expected_returncode, (
        f'{stdout.decode("utf-8")}\n{stderr.decode("utf-8")}'
    )


@pytest.mark.parametrize('dirname', os.listdir(EXAMPLES_DIR))
def test_examples_api(dirname):
    example_dir = os.path.join(EXAMPLES_DIR, dirname)
    config_file = os.path.join(example_dir, 'mkdocs.yml')
    expected_to_raise_exc = (
        config_is_using_cache_setting(config_file) and
        not is_platformdirs_installed()
    )

    def run():
        cfg = config.load_config(config_file=config_file)
        build(cfg, dirty=False)

    if expected_to_raise_exc:
        with pytest.raises(Abort):
            run()
    else:
        run()



"""Module where the `on_page_markdown` plugin event is processed."""

from __future__ import annotations

import functools
import html
import os
import re
import textwrap
from dataclasses import dataclass
from typing import TYPE_CHECKING

from mkdocs.exceptions import PluginError
from wcmatch import glob

from mkdocs_include_markdown_plugin import process
from mkdocs_include_markdown_plugin.cache import Cache
from mkdocs_include_markdown_plugin.directive import (
    ARGUMENT_REGEXES,
    GLOB_FLAGS,
    create_include_tag,
    parse_bool_options,
    parse_filename_argument,
    parse_string_argument,
    resolve_file_paths_to_exclude,
    resolve_file_paths_to_include,
    validate_order_option,
    warn_invalid_directive_arguments,
)
from mkdocs_include_markdown_plugin.files_watcher import FilesWatcher
from mkdocs_include_markdown_plugin.logger import logger
from mkdocs_include_markdown_plugin.placeholders import (
    escape_placeholders,
    save_placeholder,
    unescape_placeholders,
)


if TYPE_CHECKING:  # pragma: no cover
    from typing import TypedDict

    from mkdocs.structure.pages import Page

    from mkdocs_include_markdown_plugin.directive import DefaultValues
    from mkdocs_include_markdown_plugin.plugin import IncludeMarkdownPlugin

    IncludeTags = TypedDict(
        'IncludeTags', {
            'include': re.Pattern[str],
            'include-markdown': re.Pattern[str],
        },
    )


@dataclass
class Settings:  # noqa: D101
    exclude: list[str] | None


def get_file_content(  # noqa: PLR0913, PLR0915
        markdown: str,
        # Generated pages return `None` for `file.abs_src_path` because
        # they are not read from a file. In this case, page_src_path is
        # set to `None`.
        page_src_path: str | None,
        docs_dir: str,
        tags: IncludeTags,
        defaults: DefaultValues,
        settings: Settings,
        cumulative_heading_offset: int = 0,
        files_watcher: FilesWatcher | None = None,
        http_cache: Cache | None = None,
) -> str:
    """Return the content of the file to include."""
    if settings.exclude:
        settings_ignore_paths = list(glob.glob(
            [
                os.path.join(docs_dir, fp)
                if not os.path.isabs(fp)
                else fp for fp in settings.exclude
            ],
            flags=GLOB_FLAGS,
            root_dir=docs_dir,
        ))
        if page_src_path in settings_ignore_paths:
            return markdown
    else:
        settings_ignore_paths = []

    markdown = escape_placeholders(markdown)
    placeholders_contents: list[tuple[str, str]] = []

    def found_include_tag(  # noqa: PLR0912, PLR0915
            match: re.Match[str],
    ) -> str:
        directive_match_start = match.start()
        directive_lineno = functools.partial(
            process.lineno_from_content_start,
            markdown,
            directive_match_start,
        )

        includer_indent = match['_includer_indent']

        filename, raw_filename = parse_filename_argument(match)
        if filename is None:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                "Found no path passed including with 'include'"
                f' directive at {location}',
            )

        arguments_string = match['arguments']

        used_arguments = warn_invalid_directive_arguments(
            arguments_string,
            directive_lineno,
            'include',
            page_src_path,
            docs_dir,
        )

        ignore_paths = [*settings_ignore_paths]
        if 'exclude' in used_arguments:
            exclude_match = ARGUMENT_REGEXES['exclude']().search(
                arguments_string,
            )
            exclude_string = parse_string_argument(exclude_match)
            if exclude_string is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'exclude' argument in 'include'"
                    f' directive at {location}',
                )

            for path in resolve_file_paths_to_exclude(
                exclude_string, page_src_path, docs_dir,
            ):
                ignore_paths.append(path)

        order = defaults['order']
        if 'order' in used_arguments:
            order_match = ARGUMENT_REGEXES['order']().search(
                arguments_string,
            )
            order_ = parse_string_argument(order_match)
            if order_ is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'order' argument in 'include'"
                    f' directive at {location}',
                )
            validate_order_option(
                order_, page_src_path, docs_dir, directive_lineno, 'include',
            )
            order = order_

        file_paths_to_include, is_url = resolve_file_paths_to_include(
            filename,
            page_src_path,
            docs_dir,
            ignore_paths,
            order,
        )

        if is_url and 'order' in used_arguments:  # pragma: no cover
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            logger.warning(
                f"Ignoring 'order' argument of 'include' directive"
                f" at {location} because the included path is a URL",
            )

        if not file_paths_to_include:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                f"No files found including '{raw_filename}' at {location}",
            )

        if files_watcher is not None and not is_url:
            files_watcher.included_files.extend(file_paths_to_include)

        start = defaults['start']
        if 'start' in used_arguments:
            start_match = ARGUMENT_REGEXES['start']().search(arguments_string)
            start = parse_string_argument(start_match)
            if start is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'start' argument in 'include' directive"
                    f' at {location}',
                )

        end = defaults['end']
        if 'end' in used_arguments:
            end_match = ARGUMENT_REGEXES['end']().search(arguments_string)
            end = parse_string_argument(end_match)
            if end is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'end' argument in 'include' directive at"
                    f' {location}',
                )

        encoding = defaults['encoding']
        if 'encoding' in used_arguments:
            encoding_match = ARGUMENT_REGEXES['encoding']().search(
                arguments_string,
            )
            encoding_ = parse_string_argument(encoding_match)
            if encoding_ is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'encoding' argument in 'include'"
                    f' directive at {location}',
                )
            encoding = encoding_

        bool_options, invalid_bool_args = parse_bool_options(
            [
                'preserve-includer-indent',
                'dedent',
                'trailing-newlines',
                'recursive',
            ],
            defaults,
            arguments_string,
            used_arguments,
        )
        if invalid_bool_args:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                f"Invalid value for '{invalid_bool_args[0]}' argument of"
                f" 'include' directive at {location}."
                f' Possible values are true or false.',
            )

        text_to_include = ''
        expected_but_any_found = [start is not None, end is not None]
        for file_path in file_paths_to_include:
            if process.is_url(filename):
                new_text_to_include = process.read_url(
                    file_path, http_cache, encoding,
                )
            else:
                new_text_to_include = process.read_file(file_path, encoding)

            if start or end:
                new_text_to_include, *expected_not_found = (
                    process.filter_inclusions(
                        start,
                        end,
                        new_text_to_include,
                    )
                )
                for i in range(2):
                    if expected_but_any_found[i] and not expected_not_found[i]:
                        expected_but_any_found[i] = False

            # nested includes
            if bool_options['recursive'].value:
                new_text_to_include = get_file_content(
                    new_text_to_include,
                    file_path,
                    docs_dir,
                    tags,
                    defaults,
                    settings,
                    files_watcher=files_watcher,
                    http_cache=http_cache,
                )

            # trailing newlines right stripping
            if not bool_options['trailing-newlines'].value:
                new_text_to_include = process.rstrip_trailing_newlines(
                    new_text_to_include,
                )

            if bool_options['dedent'].value:
                new_text_to_include = textwrap.dedent(new_text_to_include)

            # includer indentation preservation
            if bool_options['preserve-includer-indent'].value:
                new_text_to_include = ''.join(
                    includer_indent + line
                    for line in (
                        new_text_to_include.splitlines(keepends=True)
                        or ['']
                    )
                )
            else:
                new_text_to_include = includer_indent + new_text_to_include

            text_to_include += new_text_to_include

        # warn if expected start or ends haven't been found in included content
        for i, delimiter_name in enumerate(['start', 'end']):
            if expected_but_any_found[i]:
                delimiter_value = locals()[delimiter_name]
                readable_files_to_include = ', '.join([
                    process.safe_os_path_relpath(fpath, docs_dir)
                    for fpath in file_paths_to_include
                ])
                plural_suffix = 's' if len(file_paths_to_include) > 1 else ''
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                logger.warning(
                    f"Delimiter {delimiter_name} '{delimiter_value}'"
                    f" of 'include' directive at {location}"
                    f' not detected in the file{plural_suffix}'
                    f' {readable_files_to_include}',
                )

        return save_placeholder(placeholders_contents, text_to_include)

    def found_include_markdown_tag(  # noqa: PLR0912, PLR0915
            match: re.Match[str],
    ) -> str:
        directive_match_start = match.start()
        directive_lineno = functools.partial(
            process.lineno_from_content_start,
            markdown,
            directive_match_start,
        )

        includer_indent = match['_includer_indent']
        filled_includer_indent = ' ' * len(includer_indent)

        filename, raw_filename = parse_filename_argument(match)
        if filename is None:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                "Found no path passed including with 'include-markdown'"
                f' directive at {location}',
            )

        arguments_string = match['arguments']

        used_arguments = warn_invalid_directive_arguments(
            arguments_string,
            directive_lineno,
            'include-markdown',
            page_src_path,
            docs_dir,
        )

        ignore_paths = [*settings_ignore_paths]
        if 'exclude' in used_arguments:
            exclude_match = ARGUMENT_REGEXES['exclude']().search(
                arguments_string,
            )
            exclude_string = parse_string_argument(exclude_match)
            if exclude_string is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'exclude' argument in 'include-markdown'"
                    f' directive at {location}',
                )
            for path in resolve_file_paths_to_exclude(
                exclude_string, page_src_path, docs_dir,
            ):
                ignore_paths.append(path)

        order = defaults['order']
        if 'order' in used_arguments:
            order_match = ARGUMENT_REGEXES['order']().search(
                arguments_string,
            )
            order_ = parse_string_argument(order_match)
            if order_ is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'order' argument in 'include-markdown'"
                    f' directive at {location}',
                )
            validate_order_option(
                order_,
                page_src_path,
                docs_dir,
                directive_lineno,
                'include-markdown',
            )
            order = order_

        file_paths_to_include, is_url = resolve_file_paths_to_include(
            filename,
            page_src_path,
            docs_dir,
            ignore_paths,
            order,
        )

        if is_url and 'order' in used_arguments:  # pragma: no cover
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            logger.warning(
                f"Ignoring 'order' argument of 'include-markdown' directive"
                f" at {location} because the included path is a URL",
            )

        if not file_paths_to_include:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                f"No files found including '{raw_filename}' at {location}",
            )

        if files_watcher is not None and not is_url:
            files_watcher.included_files.extend(file_paths_to_include)

        # start and end arguments
        start = defaults['start']
        if 'start' in used_arguments:
            start_match = ARGUMENT_REGEXES['start']().search(arguments_string)
            start = parse_string_argument(start_match)
            if start is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'start' argument in"
                    f" 'include-markdown' directive at {location}",
                )

        end = defaults['end']
        if 'end' in used_arguments:
            end_match = ARGUMENT_REGEXES['end']().search(arguments_string)
            end = parse_string_argument(end_match)
            if end is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'end' argument in 'include-markdown'"
                    f' directive at {location}',
                )

        encoding = defaults['encoding']
        if 'encoding' in used_arguments:
            encoding_match = ARGUMENT_REGEXES['encoding']().search(
                arguments_string,
            )
            encoding_ = parse_string_argument(encoding_match)
            if encoding_ is None:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'encoding' argument in"
                    f" 'include-markdown' directive at {location}",
                )
            encoding = encoding_

        # heading offset
        offset = defaults['heading-offset']
        if 'heading-offset' in used_arguments:
            offset_match = ARGUMENT_REGEXES['heading-offset']().search(
                arguments_string,
            )
            try:
                # Here None[1] would raise a TypeError
                offset_raw_value = offset_match[1]  # type: ignore
            except (IndexError, TypeError):  # pragma: no cover
                offset_raw_value = ''
            if offset_raw_value == '':
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    "Invalid empty 'heading-offset' argument in"
                    f" 'include-markdown' directive at {location}",
                )
            try:
                offset = int(offset_raw_value)
            except ValueError:
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                raise PluginError(
                    f"Invalid 'heading-offset' argument"
                    f" '{offset_raw_value}' in 'include-markdown'"
                    f" directive at {location}",
                ) from None

        bool_options, invalid_bool_args = parse_bool_options(
            [
                'rewrite-relative-urls',
                'comments',
                'preserve-includer-indent',
                'dedent',
                'trailing-newlines',
                'recursive',
            ],
            defaults,
            arguments_string,
            used_arguments,
        )
        if invalid_bool_args:
            location = process.file_lineno_message(
                page_src_path, docs_dir, directive_lineno(),
            )
            raise PluginError(
                f"Invalid value for '{invalid_bool_args[0]}' argument of"
                " 'include-markdown' directive at"
                f' {location}. Possible values are true or false.',
            )

        separator = '\n' if bool_options['trailing-newlines'].value else ''
        if not start and not end:
            start_end_part = ''
        else:
            start_end_part = f"'{html.escape(start)}' " if start else "'' "
            start_end_part += f"'{html.escape(end)}' " if end else "'' "

        # if any start or end strings are found in the included content
        # but the arguments are specified, we must raise a warning
        #
        # `True` means that no start/end strings have been found in content
        # but they have been specified, so the warning(s) must be raised
        expected_but_any_found = [start is not None, end is not None]

        text_to_include_parts = []
        for file_path in file_paths_to_include:
            if process.is_url(filename):
                new_text_to_include = process.read_url(
                    file_path, http_cache, encoding,
                )
            else:
                new_text_to_include = process.read_file(file_path, encoding)

            if start or end:
                new_text_to_include, *expected_not_found = (
                    process.filter_inclusions(
                        start,
                        end,
                        new_text_to_include,
                    )
                )
                for i in range(2):
                    if expected_but_any_found[i] and not expected_not_found[i]:
                        expected_but_any_found[i] = False

            # nested includes
            if bool_options['recursive'].value:
                new_text_to_include = get_file_content(
                    new_text_to_include,
                    file_path,
                    docs_dir,
                    tags,
                    defaults,
                    settings,
                    files_watcher=files_watcher,
                    http_cache=http_cache,
                )

            # trailing newlines right stripping
            if not bool_options['trailing-newlines'].value:
                new_text_to_include = process.rstrip_trailing_newlines(
                    new_text_to_include,
                )

            # relative URLs rewriting
            if bool_options['rewrite-relative-urls'].value:
                if page_src_path is None:  # pragma: no cover
                    logger.warning(
                        'Relative URLs rewriting is not supported in'
                        ' generated pages.',
                    )
                else:
                    new_text_to_include = process.rewrite_relative_urls(
                        new_text_to_include,
                        source_path=file_path,
                        destination_path=page_src_path,
                    )

            # comments
            if bool_options['comments'].value:
                new_text_to_include = (
                    f'{includer_indent}'
                    f'<!-- BEGIN INCLUDE {html.escape(filename)}'
                    f' {start_end_part}-->{separator}{new_text_to_include}'
                    f'{separator}<!-- END INCLUDE -->'
                )
            else:
                new_text_to_include = (
                    f'{includer_indent}{new_text_to_include}'
                )

            # dedent
            if bool_options['dedent'].value:
                new_text_to_include = textwrap.dedent(new_text_to_include)

            # includer indentation preservation
            if bool_options['preserve-includer-indent'].value and (
                new_text_to_include
            ):
                lines = new_text_to_include.splitlines(keepends=True)
                indented_lines = [lines[0]]
                for i in range(1, len(lines)):
                    indented_lines.append(filled_includer_indent + lines[i])
                new_text_to_include = ''.join(indented_lines)

            if offset:
                new_text_to_include = process.increase_headings_offset(
                    new_text_to_include,
                    offset=offset + cumulative_heading_offset,
                )

            text_to_include_parts.append(new_text_to_include)

        text_to_include = ''.join(text_to_include_parts)

        # warn if expected start or ends haven't been found in included content
        for i, delimiter_name in enumerate(['start', 'end']):
            if expected_but_any_found[i]:
                delimiter_value = locals()[delimiter_name]
                readable_files_to_include = ', '.join([
                    process.safe_os_path_relpath(fpath, docs_dir)
                    for fpath in file_paths_to_include
                ])
                plural_suffix = 's' if len(file_paths_to_include) > 1 else ''
                location = process.file_lineno_message(
                    page_src_path, docs_dir, directive_lineno(),
                )
                logger.warning(
                    f"Delimiter {delimiter_name} '{delimiter_value}' of"
                    f" 'include-markdown' directive at {location}"
                    f' not detected in the file{plural_suffix}'
                    f' {readable_files_to_include}',
                )

        return save_placeholder(placeholders_contents, text_to_include)

    # Replace contents by placeholders
    markdown = tags['include-markdown'].sub(
        found_include_markdown_tag,
        markdown,
    )
    markdown = tags['include'].sub(
        found_include_tag,
        markdown,
    )

    # Replace placeholders by contents
    for placeholder, text in placeholders_contents:
        markdown = markdown.replace(placeholder, text, 1)
    return unescape_placeholders(markdown)


def on_page_markdown(
        markdown: str,
        page: Page,
        docs_dir: str,
        plugin: IncludeMarkdownPlugin,
        http_cache: Cache | None = None,
) -> str:
    """Process markdown content of a page."""
    config = plugin.config
    return get_file_content(
        markdown,
        page.file.abs_src_path,
        docs_dir,
        {
            'include': create_include_tag(
                config.opening_tag,
                config.closing_tag,
                config.directives.get('include', 'include'),
            ),
            'include-markdown': create_include_tag(
                config.opening_tag,
                config.closing_tag,
                config.directives.get('include-markdown', 'include-markdown'),
            ),
        },
        {
            'encoding': config.encoding,
            'preserve-includer-indent': config.preserve_includer_indent,
            'dedent': config.dedent,
            'trailing-newlines': config.trailing_newlines,
            'comments': config.comments,
            'rewrite-relative-urls': config.rewrite_relative_urls,
            'heading-offset': config.heading_offset,
            'recursive': config.recursive,
            'start': config.start,
            'end': config.end,
            'order': config.order,
        },
        Settings(
            exclude=config.exclude,
        ),
        files_watcher=plugin._files_watcher,
        http_cache=plugin._cache or http_cache,
    )



import os
import sys
from dataclasses import dataclass, field

import pytest

from mkdocs_include_markdown_plugin.config import PluginConfig


parametrize_directives = pytest.mark.parametrize(
    'directive',
    ('include', 'include-markdown'),
    ids=('directive=include', 'directive=include-markdown'),
)

unix_only = pytest.mark.skipif(
    sys.platform.startswith('win'),
    reason='Test only supported on Unix systems',
)

windows_only = pytest.mark.skipif(
    not sys.platform.startswith('win'),
    reason='Test only supported on Windows systems',
)

rootdir = os.path.join(os.path.dirname(__file__), '..')


@dataclass
class FakeConfig:
    cache: int = PluginConfig.cache.default
    cache_dir: str = PluginConfig.cache_dir.default
    directives: dict[str, str] = field(
        default_factory=lambda: PluginConfig.directives.default,
    )
    order: str = PluginConfig.order.default



"""Tests for mkdocs-include-markdown-plugin `exclude` setting."""

import functools
import os
import re

import pytest
from mkdocs.exceptions import PluginError

from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives, unix_only


@unix_only
@parametrize_directives
@pytest.mark.parametrize(
    ('filenames', 'exclude', 'exclude_prefix', 'expected_result'),
    (
        pytest.param(
            ('foo', 'bar', 'baz'),
            f'content{os.sep}foo*',
            True,
            'bar\nbaz\n',
            id='ignore-by-glob',
        ),
        pytest.param(
            ('foo', 'bar', 'baz'),
            f'content{os.sep}ba*',
            True,
            'foo\n',
            id='ignore-multiple-by-glob',
        ),
        pytest.param(
            ('foo', 'bar', 'baz'),
            '',
            True,
            'bar\nbaz\nfoo\n',
            id='not-ignore',
        ),
        pytest.param(
            ('foo', 'bar', 'baz'),
            '*',
            True,
            None,
            id='ignore-all',
        ),
        pytest.param(
            ('foo', 'bar', 'baz'),
            f'..{os.sep}content{os.sep}*',
            False,
            None,
            id='ignore-all-relative',
        ),
        pytest.param(
            ('foo', 'bar', 'baz'),
            f'..{os.sep}content{os.sep}b*',
            False,
            'foo\n',
            id='ignore-relative',
        ),
    ),
)
def test_exclude(
    page,
    tmp_path,
    plugin,
    caplog,
    directive,
    filenames,
    exclude,
    exclude_prefix,
    expected_result,
):
    includer_folder = tmp_path / 'includer'
    content_folder = tmp_path / 'content'
    includer_folder.mkdir()
    content_folder.mkdir()

    includer_file = includer_folder / 'main.txt'
    files = {
        filename: content_folder / filename for filename in filenames
    }

    exclude_prefix = f'{tmp_path}{os.sep}' if exclude_prefix else ''
    includer_file_content = f'''{{%
  {directive} "{tmp_path}{os.sep}content/*"
  exclude='{exclude_prefix}{exclude}'
%}}'''
    for basename, file in files.items():
        file.write_text(f'{basename}\n')

    includer_file.write_text(includer_file_content)

    func = functools.partial(
        on_page_markdown,
        includer_file_content,
        page(includer_file),
        includer_folder,
        plugin,
    )

    if expected_result is None:
        with pytest.raises(PluginError) as exc:
            func()
        assert re.match(r'No files found including ', str(exc.value))
    else:
        assert func() == expected_result
    assert len(caplog.records) == 0



import os
import time

import pytest
from mkdocs.exceptions import PluginError

from mkdocs_include_markdown_plugin.directive import get_order_option_regex
from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives, unix_only, windows_only


@parametrize_directives
def test_default_order(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'hincluded.md'
    page_to_include_file.write_text('hincluded.md\n')
    page_to_include_file = tmp_path / 'included.md'
    page_to_include_file.write_text('included.md\n')

    assert on_page_markdown(
        f'''{{% {directive} "*.md" %}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'hincluded.md\nincluded.md\n'


@parametrize_directives
def test_default_reverse_order(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'hincluded.md'
    page_to_include_file.write_text('hincluded.md\n')
    page_to_include_file = tmp_path / 'included.md'
    page_to_include_file.write_text('included.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'included.md\nhincluded.md\n'


@parametrize_directives
def test_natural_order(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'file1.md'
    page_to_include_file.write_text('file1.md\n')
    page_to_include_file = tmp_path / 'file10.md'
    page_to_include_file.write_text('file10.md\n')
    page_to_include_file = tmp_path / 'file2.md'
    page_to_include_file.write_text('file2.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='natural'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.md\nfile2.md\nfile10.md\n'


@parametrize_directives
def test_natural_reverse_order(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'file1.md'
    page_to_include_file.write_text('file1.md\n')
    page_to_include_file = tmp_path / 'file10.md'
    page_to_include_file.write_text('file10.md\n')
    page_to_include_file = tmp_path / 'file2.md'
    page_to_include_file.write_text('file2.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-natural'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file10.md\nfile2.md\nfile1.md\n'


@parametrize_directives
@pytest.mark.parametrize('order_value', ('alpha', 'path'))
def test_alpha_order(directive, order_value, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'c.md'
    f2.write_text('c.md\n')
    f3 = tmp_path / 'b.md'
    f3.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='{order_value}'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'a.md\nb.md\nc.md\n'


@parametrize_directives
def test_alpha_reverse_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'c.md'
    f2.write_text('c.md\n')
    f3 = tmp_path / 'b.md'
    f3.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-alpha'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'c.md\nb.md\na.md\n'


@parametrize_directives
def test_random_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'b.md'
    f2.write_text('b.md\n')
    f3 = tmp_path / 'c.md'
    f3.write_text('c.md\n')

    result = on_page_markdown(
        f'''{{%
{directive} "*.md"
order='random'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ).splitlines()

    assert set(result) == {'a.md', 'b.md', 'c.md'}


@parametrize_directives
def test_system_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'b.md'
    f2.write_text('b.md\n')

    result = on_page_markdown(
        f'''{{%
{directive} "*.md"
order='system'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ).splitlines()

    assert set(result) == {'a.md', 'b.md'}


@parametrize_directives
def test_size_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'small.md'
    small_content = 'a' * 10
    f1.write_text(small_content)
    f2 = tmp_path / 'large.md'
    large_content = 'b' * 100
    f2.write_text(large_content)

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='size'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == large_content + small_content


@parametrize_directives
def test_size_reverse_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'small.md'
    small_content = 'a' * 10
    f1.write_text(small_content)
    f2 = tmp_path / 'large.md'
    large_content = 'b' * 100
    f2.write_text(large_content)

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-size'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == small_content + large_content


@parametrize_directives
def test_mtime_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'older.md'
    f1.write_text('older.md\n')
    f2 = tmp_path / 'newer.md'
    f2.write_text('newer.md\n')
    now = time.time()
    os.utime(f1, (now - 10, now - 10))
    os.utime(f2, (now, now))
    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='mtime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'older.md\nnewer.md\n'


@parametrize_directives
def test_mtime_reverse_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'older.md'
    f1.write_text('older.md\n')
    f2 = tmp_path / 'newer.md'
    f2.write_text('newer.md\n')
    now = time.time()
    os.utime(f1, (now - 10, now - 10))
    os.utime(f2, (now, now))
    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-mtime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'newer.md\nolder.md\n'


@unix_only
@parametrize_directives
def test_atime_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'older.md'
    f1.write_text('older.md\n')
    f2 = tmp_path / 'newer.md'
    f2.write_text('newer.md\n')
    os.utime(f1, (time.time() - 10, time.time() - 10))

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='atime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'older.md\nnewer.md\n'


@unix_only
@parametrize_directives
def test_atime_reverse_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'older.md'
    f1.write_text('older.md\n')
    f2 = tmp_path / 'newer.md'
    f2.write_text('newer.md\n')
    os.utime(f1, (time.time() - 10, time.time() - 10))

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-atime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'newer.md\nolder.md\n'


@parametrize_directives
def test_alpha_order_by_path(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/a.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/a.md\n')
    f2 = tmp_path / 'b.md'
    f2.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='alpha-path'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'b.md\nsub/a.md\n'


@parametrize_directives
@pytest.mark.parametrize('order_value', ('alpha-extension', 'extension'))
def test_alpha_order_by_extension(
    directive, order_value, page, tmp_path, plugin,
):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    f2 = tmp_path / 'file1.txt'
    f2.write_text('file1.txt\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*"
order='{order_value}'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file2.md\nfile1.txt\n'


@parametrize_directives
def test_natural_order_by_path(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/file2.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/file2.md\n')
    f2 = tmp_path / 'file10.md'
    f2.write_text('file10.md\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='natural-path'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.md\nfile10.md\nsub/file2.md\n'


@parametrize_directives
def test_natural_order_by_name(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/file2.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/file2.md\n')
    f2 = tmp_path / 'file10.md'
    f2.write_text('file10.md\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='natural-name'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.md\nsub/file2.md\nfile10.md\n'


@parametrize_directives
def test_natural_order_by_extension(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    f2 = tmp_path / 'file10.txt'
    f2.write_text('file10.txt\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*"
order='natural-extension'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.md\nfile2.md\nfile10.txt\n'


@parametrize_directives
@pytest.mark.parametrize('order_value', ('alpha-name', 'name'))
def test_alpha_order_by_name(directive, order_value, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'c.md'
    f2.write_text('c.md\n')
    f3 = tmp_path / 'b.md'
    f3.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='{order_value}'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'a.md\nb.md\nc.md\n'


@parametrize_directives
def test_alpha_order_by_name_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'a.md'
    f1.write_text('a.md\n')
    f2 = tmp_path / 'c.md'
    f2.write_text('c.md\n')
    f3 = tmp_path / 'b.md'
    f3.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-alpha-name'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'c.md\nb.md\na.md\n'


@parametrize_directives
def test_alpha_order_by_path_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/a.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/a.md\n')
    f2 = tmp_path / 'b.md'
    f2.write_text('b.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='-alpha-path'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'sub/a.md\nb.md\n'


@parametrize_directives
def test_alpha_order_by_extension_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    f2 = tmp_path / 'file1.txt'
    f2.write_text('file1.txt\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*"
order='-alpha-extension'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.txt\nfile2.md\n'


@parametrize_directives
def test_natural_order_by_name_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/file2.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/file2.md\n')
    f2 = tmp_path / 'file10.md'
    f2.write_text('file10.md\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='-natural-name'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file10.md\nsub/file2.md\nfile1.md\n'


@parametrize_directives
def test_natural_order_by_path_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'sub/file2.md'
    f1.parent.mkdir(parents=True)
    f1.write_text('sub/file2.md\n')
    f2 = tmp_path / 'file10.md'
    f2.write_text('file10.md\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "./**/*.md"
order='-natural-path'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'sub/file2.md\nfile10.md\nfile1.md\n'


@parametrize_directives
def test_natural_order_by_extension_reverse(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    f2 = tmp_path / 'file10.txt'
    f2.write_text('file10.txt\n')
    f3 = tmp_path / 'file1.md'
    f3.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*"
order='-natural-extension'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file10.txt\nfile1.md\nfile2.md\n'


@parametrize_directives
@windows_only
def test_ctime_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    time.sleep(1)
    f2 = tmp_path / 'file1.md'
    f2.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='ctime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file2.md\nfile1.md\n'


@windows_only
@parametrize_directives
def test_ctime_reverse_order(directive, page, tmp_path, plugin):
    f1 = tmp_path / 'file2.md'
    f1.write_text('file2.md\n')
    time.sleep(1)
    f2 = tmp_path / 'file1.md'
    f2.write_text('file1.md\n')

    assert on_page_markdown(
        f'''{{%
{directive} "*.md"
order='-ctime'
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'file1.md\nfile2.md\n'


@unix_only
@parametrize_directives
def test_invalid_empty_order_argument(
    directive, page, tmp_path, caplog, plugin,
):
    directory_to_include = tmp_path / 'empty_order_argument'
    directory_to_include.mkdir()

    page_to_include_filepath = directory_to_include / 'included.md'
    page_to_include_filepath.write_text('Content that should be included\n')

    page_to_exclude_filepath = directory_to_include / 'igno"re"d.md'
    page_to_exclude_filepath.write_text('Content that should be excluded\n')

    includer_glob = os.path.join(str(directory_to_include), '*.md')

    includer_file_content = f'''{{%
  {directive} "{includer_glob}"
  order=
%}}'''

    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            includer_file_content,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

    assert len(caplog.records) == 0
    assert str(exc.value) == (
        f"Invalid empty 'order' argument in '{directive}' directive"
        ' at includer.md:1'
    )


@parametrize_directives
def test_invalid_order_argument(directive, page, tmp_path, caplog, plugin):
    directory_to_include = tmp_path / 'invalid_order_argument'
    directory_to_include.mkdir()

    page_to_include_filepath = directory_to_include / 'included.md'
    page_to_include_filepath.write_text('Content that should be included\n')

    page_to_exclude_filepath = directory_to_include / 'ignored.md'
    page_to_exclude_filepath.write_text('Content that should be excluded\n')

    includer_glob = os.path.join(str(directory_to_include), '*.md')

    includer_file_content = f'''{{%
  {directive} "{includer_glob}"
  order='invalid-order'
%}}'''

    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            includer_file_content,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

    assert len(caplog.records) == 0
    regex = get_order_option_regex()
    assert str(exc.value) == (
        f"Invalid value 'invalid-order' for the 'order' argument in"
        f" '{directive}' directive at includer.md:1. The argument"
        f" 'order' must be a string that matches the regex '{regex.pattern}'."
    )



import functools
import os
import re

import pytest
from mkdocs.exceptions import PluginError

from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives, unix_only


@pytest.mark.parametrize(
    ('directive', 'arguments'),
    (
        pytest.param(
            'include',
            [
                'preserve-includer-indent',
                'dedent',
                'trailing-newlines',
            ],
            id='include',
        ),
        pytest.param(
            'include-markdown',
            [
                'preserve-includer-indent',
                'dedent',
                'rewrite-relative-urls',
                'comments',
                'trailing-newlines',
            ],
            id='include-markdown',
        ),
    ),
)
def test_invalid_bool_arguments(
    directive,
    arguments,
    page,
    tmp_path,
    plugin,
    caplog,
):
    for argument_name in arguments:
        page_to_include_filepath = tmp_path / 'included.md'
        page_to_include_filepath.write_text('Included\n')

        filename = 'includer.md'

        with pytest.raises(PluginError) as exc:
            on_page_markdown(
                f'''{{%
    {directive} "{page_to_include_filepath}"
    {argument_name}=invalidoption
    %}}''',
                page(tmp_path / filename),
                tmp_path,
                plugin,
            )
        assert str(exc.value) == (
            f"Invalid value for '{argument_name}' argument of '{directive}'"
            f' directive at {filename}:1. Possible values are true or false.'
        )
        assert len(caplog.records) == 0


@parametrize_directives
def test_start_end_mixed_quotes(directive, page, caplog, tmp_path, plugin):
    page_to_include_filepath = tmp_path / 'included.md'
    page_to_include_filepath.write_text('''Content that should be ignored
<!-- "s'tar't" -->
Content to include
<!-- 'en"d' -->
More content that should be ignored
''')

    includer_file_content = f'''{{%
  {directive} "{page_to_include_filepath}"
  start='<!-- "s\\'tar\\'t" -->'
  end="<!-- 'en\\"d' -->"
%}}'''
    result = on_page_markdown(
        includer_file_content,
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    )
    assert result == '\nContent to include\n'

    assert caplog.records == []


@pytest.mark.parametrize('argument', ('start', 'end'))
@parametrize_directives
def test_invalid_start_end_arguments(
    argument,
    directive,
    page,
    caplog,
    tmp_path,
    plugin,
):
    page_to_include_filepath = tmp_path / 'included.md'
    included_content = '''Content that should be ignored
<!-- start -->
Content to include
<!-- end -->
More content that should be ignored
'''
    page_to_include_filepath.write_text(included_content)

    includer_file_content = f'''
{{%
  {directive} "{page_to_include_filepath}"
  {argument}=''
%}}'''
    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            includer_file_content,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )
    assert str(exc.value) == (
        f"Invalid empty '{argument}' argument in '{directive}'"
        ' directive at includer.md:2'
    )

    assert len([record.msg for record in caplog.records]) == 0


@unix_only
@parametrize_directives
def test_exclude_double_quote_escapes(
    directive, page, tmp_path, plugin, caplog,
):
    drectory_to_include = tmp_path / 'exclude_double_quote_escapes'
    drectory_to_include.mkdir()

    page_to_include_filepath = drectory_to_include / 'included.md'
    page_to_include_filepath.write_text('Content that should be included\n')

    page_to_exclude_filepath = drectory_to_include / 'igno"re"d.md'
    page_to_exclude_filepath.write_text('Content that should be excluded\n')
    page_to_exclude_escaped_filepath = str(
        page_to_exclude_filepath,
    ).replace('"', '\\"')

    includer_glob = os.path.join(str(drectory_to_include), '*.md')
    includer_file_content = f'''{{%
  {directive} "{includer_glob}"
  exclude="{page_to_exclude_escaped_filepath}"
%}}'''
    result = on_page_markdown(
        includer_file_content,
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    )
    assert result == 'Content that should be included\n'
    assert caplog.records == []


@unix_only
@parametrize_directives
def test_invalid_empty_exclude_argument(
    directive, page, tmp_path, caplog, plugin,
):
    directory_to_include = tmp_path / 'exclude_double_quote_escapes'
    directory_to_include.mkdir()

    page_to_include_filepath = directory_to_include / 'included.md'
    page_to_include_filepath.write_text('Content that should be included\n')

    page_to_exclude_filepath = directory_to_include / 'igno"re"d.md'
    page_to_exclude_filepath.write_text('Content that should be excluded\n')

    includer_glob = os.path.join(str(directory_to_include), '*.md')

    includer_file_content = f'''{{%
  {directive} "{includer_glob}"
  exclude=
%}}'''

    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            includer_file_content,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

    assert len(caplog.records) == 0
    assert str(exc.value) == (
        f"Invalid empty 'exclude' argument in '{directive}' directive"
        ' at includer.md:1'
    )


@parametrize_directives
def test_invalid_empty_encoding_argument(
    directive, page, tmp_path, plugin, caplog,
):
    page_to_include_filepath = tmp_path / 'included.md'
    page_to_include_filepath.write_text('Content to include')

    includer_file_content = f'''{{%
  {directive} "{page_to_include_filepath}"
  encoding=
%}}'''

    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            includer_file_content,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

    assert len(caplog.records) == 0
    assert str(exc.value) == (
        f"Invalid empty 'encoding' argument in '{directive}'"
        ' directive at includer.md:1'
    )


@pytest.mark.parametrize(
    ('argument_value', 'exception_message'),
    (
        pytest.param(
            'invalidoption', (
                "Invalid 'heading-offset' argument 'invalidoption' in"
                " 'include-markdown' directive at includer.md:1"
            ),
            id='invalidoption',
        ),
        pytest.param(
            '', (
                "Invalid empty 'heading-offset' argument in"
                " 'include-markdown' directive at includer.md:1"
            ),
            id='empty',
        ),
    ),
)
def test_invalid_heading_offset_arguments(
    argument_value,
    exception_message,
    page,
    tmp_path,
    plugin,
    caplog,
):
    page_to_include_filepath = tmp_path / 'included.md'
    page_to_include_filepath.write_text('# Content to include')

    with pytest.raises(PluginError) as exc:
        on_page_markdown(
            f'''{{%
  include-markdown "{page_to_include_filepath}"
  heading-offset={argument_value}
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

    assert len(caplog.records) == 0
    assert str(exc.value) == exception_message


@parametrize_directives
def test_invalid_argument_name(directive, page, tmp_path, plugin, caplog):
    page_to_include_filepath = tmp_path / 'included.md'
    page_to_include_filepath.write_text('Content to include')

    includer_file_content = f'''{{%
  {directive} "{page_to_include_filepath}"
  invalid-argument=true
%}}'''
    assert on_page_markdown(
        includer_file_content,
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    ) == 'Content to include'

    assert len(caplog.records) == 1
    assert caplog.records[0].msg == (
        f"Invalid argument 'invalid-argument' in '{directive}'"
        " directive at includer.md:1. Ignoring..."
    )


class TestFilename:
    double_quoted_filenames = [
        'inc"luded.md', 'inc"lude"d.md', 'included.md"', '"included.md',
    ]
    single_quoted_filenames = [
        fname.replace('"', "'") for fname in double_quoted_filenames
    ]

    @unix_only
    @parametrize_directives
    @pytest.mark.parametrize('filename', double_quoted_filenames)
    def test_not_escaped_double_quotes(
        self, directive, filename, page, tmp_path, plugin, caplog,
    ):
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text('Foo\n')

        with pytest.raises(PluginError) as exc:
            on_page_markdown(
                f'{{% {directive} "{page_to_include_filepath}" %}}',
                page(tmp_path / 'includer.md'),
                tmp_path,
                plugin,
            )

        assert len(caplog.records) == 0
        assert re.match(
            r'^No files found including ',
            str(exc.value),
        )

    @unix_only
    @parametrize_directives
    @pytest.mark.parametrize('filename', double_quoted_filenames)
    def test_escaped_double_quotes(
        self, directive, filename, page, tmp_path, plugin,
    ):
        included_content = 'Foo\n'
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text(included_content)

        # escape filename passed as argument
        escaped_page_to_include_filepath = str(
            page_to_include_filepath,
        ).replace('"', '\\"')
        result = on_page_markdown(
            f'''{{%
  {directive} "{escaped_page_to_include_filepath}"
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )
        assert result == included_content

    @parametrize_directives
    @pytest.mark.parametrize('filename', single_quoted_filenames)
    def test_escaped_single_quotes(
        self, filename, directive, page, tmp_path, plugin,
    ):
        included_content = 'Foo\n'
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text(included_content)

        # escape filename passed as argument
        escaped_page_to_include_filepath = str(
            page_to_include_filepath,
        ).replace("'", "\\'")
        result = on_page_markdown(
            f'''{{%
  {directive} '{escaped_page_to_include_filepath}'
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )
        assert result == included_content

    @unix_only
    @parametrize_directives
    @pytest.mark.parametrize('filename', double_quoted_filenames)
    def test_unescaped_double_quotes(
        self, filename, directive, page, tmp_path, plugin,
    ):
        included_content = 'Foo\n'
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text(included_content)

        result = on_page_markdown(
            f'''{{%
  {directive} '{page_to_include_filepath}'
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )
        assert result == included_content

    @parametrize_directives
    @pytest.mark.parametrize('filename', single_quoted_filenames)
    def test_unescaped_single_quotes(
        self, filename, directive, page, tmp_path, plugin,
    ):
        included_content = 'Foo\n'
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text(included_content)

        result = on_page_markdown(
            f'''{{%
  {directive} "{page_to_include_filepath}"
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )
        assert result == included_content

    @unix_only
    @parametrize_directives
    @pytest.mark.parametrize(
        'filename', ("inc'luded\".md", "''i\"nc\"lude'd.md"),
    )
    @pytest.mark.parametrize(
        'quote', ('"', "'"), ids=('quote="', "quote='"),
    )
    @pytest.mark.parametrize(
        'escape', (True, False), ids=('escape=True', 'escape=False'),
    )
    def test_mixed_quotes(
        self,
        filename,
        quote,
        escape,
        directive,
        page,
        tmp_path,
        plugin,
        caplog,
    ):
        included_content = 'Foo\n'
        page_to_include_filepath = tmp_path / filename
        page_to_include_filepath.write_text(included_content)

        if escape:
            escaped_page_to_include_filepath = str(
                page_to_include_filepath,
            ).replace(quote, f'\\{quote}')
        else:
            escaped_page_to_include_filepath = page_to_include_filepath

        markdown = f'''{{%
  {directive} {quote}{escaped_page_to_include_filepath}{quote}
%}}'''

        func = functools.partial(
            on_page_markdown,
            markdown,
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )

        if escape:
            assert func() == included_content
        else:
            with pytest.raises(PluginError) as exc:
                func()
            assert re.match(
                r'No files found including ',
                str(exc.value),
            )
        assert len(caplog.records) == 0

    @parametrize_directives
    def test_no_filename(self, directive, page, tmp_path, plugin, caplog):
        filename = 'includer.md'

        with pytest.raises(PluginError) as exc:
            on_page_markdown(
                f'\n\n{{% {directive} %}}',
                page(tmp_path / filename),
                tmp_path,
                plugin,
            )

        assert str(exc.value) == (
            f"Found no path passed including with '{directive}' directive"
            f' at {filename}:3'
        )
        assert len(caplog.records) == 0

    @parametrize_directives
    def test_non_existent_filename(
        self,
        directive,
        page,
        tmp_path,
        plugin,
        caplog,
    ):
        page_content = f'''{{%
    {directive} "/path/to/file/that/does/not/exists"
    start="<!--start-here-->"
    end="<!--end-here-->"
%}}'''

        page_filepath = tmp_path / 'example.md'
        page_filepath.write_text(page_content)

        with pytest.raises(PluginError) as exc:
            on_page_markdown(
                page_content,
                page(page_filepath),
                tmp_path,
                plugin,
            )

        assert len(caplog.records) == 0
        assert re.match(r'No files found including ', str(exc.value))



"""Module for placeholders processing."""

# Placeholders (taken from Python-Markdown)
from __future__ import annotations


STX = '\u0002'
''' "Start of Text" marker for placeholder templates. '''
ETX = '\u0003'
''' "End of Text" marker for placeholder templates. '''
INLINE_PLACEHOLDER_PREFIX = f'{STX}klzzwxh:'


def build_placeholder(num: int) -> str:
    """Return a placeholder."""
    return f'{INLINE_PLACEHOLDER_PREFIX}{num}{ETX}'


def escape_placeholders(text: str) -> str:
    """Escape placeholders in the given text."""
    return text.replace(STX, f'\\{STX}').replace(ETX, f'\\{ETX}')


def unescape_placeholders(text: str) -> str:
    """Unescape placeholders in the given text."""
    return text.replace(f'\\{STX}', STX).replace(f'\\{ETX}', ETX)


def save_placeholder(
        placeholders_contents: list[tuple[str, str]],
        text_to_include: str,
) -> str:
    """Save the included text and return the placeholder."""
    inclusion_index = len(placeholders_contents)
    placeholder = build_placeholder(inclusion_index)
    placeholders_contents.append((placeholder, text_to_include))
    return placeholder



import os
import time

from mkdocs_include_markdown_plugin.cache import (
    Cache,
    get_cache_directory,
    initialize_cache,
    is_platformdirs_installed,
)


def test_cache_read_file(tmp_path):
    cache = Cache(tmp_path)
    assert cache.read_file('pyproject.toml').split('\n', 1)[0] == (
        'name = "mkdocs-include-markdown-plugin"'
    )


def test_cache_expiration_on_get(tmp_path):
    cache = Cache(tmp_path)
    cache.set_('foo', f'{time.time() - 600*10}\nbar')
    assert cache.get_('foo') is None


def test_cache_clean(tmp_path):
    now_ts = int(time.time())

    file1 = tmp_path / 'file1'
    file1.write_text(f'{now_ts}\n')
    file2 = tmp_path / 'file2'
    file2.write_text(f'{now_ts}\n')

    assert len(os.listdir(tmp_path)) == 2

    cache = Cache(tmp_path, 0)
    cache.clean()

    assert len(os.listdir(tmp_path)) == 0


def test_get_cache_directory_empty():
    if not is_platformdirs_installed():
        assert get_cache_directory('') is None
    else:
        assert isinstance(get_cache_directory(''), str)


def test_get_cache_directory_custom():
    assert get_cache_directory('foo') == 'foo'


def test_initialize_cache_not_cache_dir():
    if not is_platformdirs_installed():
        assert initialize_cache(300, '') is None
    else:
        assert isinstance(initialize_cache(300, ''), Cache)


def test_initialize_cache_cache_dir():
    assert isinstance(initialize_cache(300, 'foo'), Cache)



"""Utilities for string processing."""

from __future__ import annotations

import functools
import io
import os
import re
import stat
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING


if TYPE_CHECKING:  # pragma: no cover
    from typing import Any

    from mkdocs_include_markdown_plugin.cache import Cache
    from mkdocs_include_markdown_plugin.directive import OrderOption


# Markdown regular expressions. Taken from the original Markdown.pl by John
# Gruber, and modified to work in Python

# Matches markdown links.
# e.g. [scikit-learn](https://github.com/scikit-learn/scikit-learn)
#
# The next Regex can raise a catastrophic backtracking, but with the current
# implementation of the plugin it is not very much likely to reach the case.
# Can be checked with dlint:
# python3 -m dlint.redos --pattern '\[(?:(?:\[[^\[\]]+\])*)?\]'
#
# In the original Markdown.pl, the nested brackets are enclosed by an atomic
# group (?>...), but atomic groups are not supported by Python in versions
# previous to Python3.11. Also, these nested brackets can be recursive in the
# Perl implementation but this doesn't seem possible in Python, the current
# implementation only reaches two levels.
MARKDOWN_LINK_REGEX = re.compile(
    r'''
        (                 # wrap whole match in $1
          (?<!!)          # don't match images - negative lookbehind
          \[
            (             # link text = $2
                (?:
                    [^\[\]]+  # not bracket
                    (?:
                        \[[^\[\]]+\]  # another level of nested bracket
                                      # with something inside
                        [^\[\]]*      # not bracket
                    )*
                )?        # allow for empty link text
            )
          \]
          \(             # literal paren
            \s*
            <?(.*?)>?    # href = $3
            \s*
            (            # $4
              (['"])     # quote char = $5
              (.*?)      # Title = $6
              \5         # matching quote
            )?           # title is optional
          \)
        )
    ''',
    flags=re.VERBOSE,
)

# Matches markdown inline images.
# e.g. ![alt-text](path/to/image.png)
MARKDOWN_IMAGE_REGEX = re.compile(
    r'''
        (                # wrap whole match in $1
          !\[
            (.*?)        # alt text = $2
          \]
          \(             # literal paren
            [ \t]*
            <?(\S+?)>?   # src url = $3
            [ \t]*
            (            # $4
              (['"])     # quote char = $5
              (.*?)      # title = $6
              \5         # matching quote
              [ \t]*
            )?           # title is optional
          \)
        )
    ''',
    flags=re.VERBOSE,
)

# Matches markdown link definitions.
# e.g. [scikit-learn]: https://github.com/scikit-learn/scikit-learn
MARKDOWN_LINK_DEFINITION_REGEX = re.compile(
    r'''
        ^[ ]{0,4}\[(.+)\]:   # id = $1
        [ \t]*
        \n?                # maybe *one* newline
        [ \t]*
        <?(\S+?)>?           # url = $2
        [ \t]*
        \n?                # maybe one newline
        [ \t]*
        (?:
            (?<=\s)          # lookbehind for whitespace
            ["(]
            (.+?)            # title = $3
            [")]
            [ \t]*
        )?                   # title is optional
        (?:\n+|\Z)
    ''',
    flags=re.VERBOSE | re.MULTILINE,
)

# Matched html image and source definition.
# e.g. <img src="path/to/image.png" alt="alt-text">
# e.g. <source src="path/to/image.png" alt="alt-text">
MARKDOWN_HTML_IMAGE_REGEX = re.compile(
    r'''
        <(?:img|source)    # img or source
        (?:\s+             # More than one whitespace
            (?!src=)       # Not src=
            [\w-]+         # attribute name
            (?:\s*=\s*)?   # arbitrary whitespace (optional)
            (?:
                "[^"]*"    # Quoted value (double quote)
                |
                '[^']*'    # Quoted value (single quote)
            )?
        )*                 # Other attributes are repeated 0 or more times
        \s+                # More than one whitespace
        src=["'](\S+?)["'] # src = $1 (double quote or single quote)
    ''',
    flags=re.VERBOSE | re.MULTILINE,
)

# Matched html anchor definition.
# e.g. <a href="https://example.com">example</a>
MARKDOWN_HTML_ANCHOR_DEFINITION_REGEX = re.compile(
    r'''
        <a
        (?:\s+             # More than one whitespace
            (?!href=)      # Not href=
            [\w-]+         # attribute name
            (?:\s*=\s*)?   # arbitrary whitespace (optional)
            (?:
                "[^"]*"    # Quoted value (double quote)
                |
                '[^']*'    # Quoted value (single quote)
            )?
        )*                 # Other attributes are repeated 0 or more times
        \s+                # More than one whitespace
        href=["'](\S+?)["']# href = $1 (double quote or single quote)
    ''',
    flags=re.VERBOSE | re.MULTILINE,
)


def transform_p_by_p_skipping_codeblocks(  # noqa: PLR0912, PLR0915
        markdown: str,
        func: Callable[[str], str],
) -> str:
    """Apply a transformation paragraph by paragraph in a Markdown text.

    Apply a transformation paragraph by paragraph in a Markdown using a
    function. Skip indented and fenced codeblock lines, where the
    transformation is never applied.
    """
    # current fenced codeblock delimiter
    _current_fcodeblock_delimiter = ''

    # inside indented codeblock
    _maybe_icodeblock_lines: list[str] = []
    _previous_line_was_empty = False

    lines = []
    current_paragraph_lines: list[str] = []

    def process_current_paragraph() -> None:
        lines.extend(func(''.join(current_paragraph_lines),
                          ).splitlines(keepends=True))

    # The next implementation takes into account that indented code
    # blocks must be surrounded by newlines as per the CommonMark
    # specification. See https://spec.commonmark.org/0.28/#indented-code-blocks
    #
    # However, note that ambiguities with list items are not handled.

    for line in io.StringIO(markdown):
        if not _current_fcodeblock_delimiter:
            lstripped_line = line.lstrip()
            if lstripped_line.startswith(('```', '~~~')):
                _current_fcodeblock_delimiter = lstripped_line[:3]
                process_current_paragraph()
                current_paragraph_lines = []
                lines.append(line)
            elif line.startswith('    '):
                if not lstripped_line or _maybe_icodeblock_lines:
                    # maybe enter indented codeblock
                    _maybe_icodeblock_lines.append(line)
                else:
                    current_paragraph_lines.append(line)
            elif _maybe_icodeblock_lines:
                process_current_paragraph()
                current_paragraph_lines = []
                if not _previous_line_was_empty:
                    # wasn't an indented code block
                    current_paragraph_lines.extend(_maybe_icodeblock_lines)
                    _maybe_icodeblock_lines = []
                    current_paragraph_lines.append(line)
                    process_current_paragraph()
                    current_paragraph_lines = []
                else:
                    # exit indented codeblock
                    for line_ in _maybe_icodeblock_lines:
                        lines.append(line_)
                    _maybe_icodeblock_lines = []
                    lines.append(line)
            else:
                current_paragraph_lines.append(line)
            _previous_line_was_empty = not lstripped_line
        else:
            lines.append(line)
            lstripped_line = line.lstrip()
            if lstripped_line.startswith(_current_fcodeblock_delimiter):
                _current_fcodeblock_delimiter = ''
            _previous_line_was_empty = not lstripped_line

    if _maybe_icodeblock_lines:
        if not _previous_line_was_empty:
            # at EOF
            process_current_paragraph()
            current_paragraph_lines = []
            current_paragraph_lines.extend(_maybe_icodeblock_lines)
            process_current_paragraph()
            current_paragraph_lines = []
        else:
            process_current_paragraph()
            current_paragraph_lines = []
            for line_ in _maybe_icodeblock_lines:
                lines.append(line_)
    else:
        process_current_paragraph()

    return ''.join(lines)


def transform_line_by_line_skipping_codeblocks(
        markdown: str,
        func: Callable[[str], str],
) -> str:
    """Apply a transformation line by line in a Markdown text using a function,.

    Skip fenced codeblock lines and empty lines, where the transformation
    is never applied.

    Indented codeblocks are not taken into account because in the practice
    this function is only used for transformations of heading prefixes. See
    the PR https://github.com/mondeja/mkdocs-include-markdown-plugin/pull/95
    to recover the implementation handling indented codeblocks.
    """
    # current fenced codeblock delimiter
    _current_fcodeblock_delimiter = ''

    lines = []
    for line in io.StringIO(markdown):
        lstripped_line = line.lstrip()
        if not _current_fcodeblock_delimiter:
            if lstripped_line.startswith('```'):
                _current_fcodeblock_delimiter = '```'
            elif lstripped_line.startswith('~~~'):
                _current_fcodeblock_delimiter = '~~~'
            else:
                line = func(line)  # noqa: PLW2901
        elif lstripped_line.startswith(_current_fcodeblock_delimiter):
            _current_fcodeblock_delimiter = ''
        lines.append(line)

    return ''.join(lines)


def rewrite_relative_urls(
        markdown: str,
        source_path: str,
        destination_path: str,
) -> str:
    """Rewrite relative URLs in a Markdown text.

    Rewrites markdown so that relative links that were written at
    ``source_path`` will still work when inserted into a file at
    ``destination_path``.
    """
    def rewrite_url(url: str) -> str:
        if is_url(url) or is_absolute_path(url) or is_anchor(url):
            return url

        new_path = os.path.relpath(
            os.path.join(os.path.dirname(source_path), url),
            os.path.dirname(destination_path),
        )

        # ensure forward slashes are used, on Windows
        new_path = new_path.replace('\\', '/').replace('//', '/')

        try:
            if url[-1] == '/':
                # the above operation removes a trailing slash,
                # so add it back if it was present in the input
                new_path += '/'
        except IndexError:  # pragma: no cover
            pass

        return new_path

    def found_href(m: re.Match[str], url_group_index: int = -1) -> str:
        match_start, match_end = m.span(0)
        href = m[url_group_index]
        href_start, href_end = m.span(url_group_index)
        rewritten_url = rewrite_url(href)
        return (
            m.string[match_start:href_start]
            + rewritten_url
            + m.string[href_end:match_end]
        )

    found_href_url_group_index_3 = functools.partial(
        found_href,
        url_group_index=3,
    )

    def transform(paragraph: str) -> str:
        paragraph = MARKDOWN_LINK_REGEX.sub(
            found_href_url_group_index_3,
            paragraph,
        )
        paragraph = MARKDOWN_IMAGE_REGEX.sub(
            found_href_url_group_index_3,
            paragraph,
        )
        paragraph = MARKDOWN_LINK_DEFINITION_REGEX.sub(
            functools.partial(found_href, url_group_index=2),
            paragraph,
        )
        paragraph = MARKDOWN_HTML_IMAGE_REGEX.sub(
            functools.partial(found_href, url_group_index=1),
            paragraph,
        )
        return MARKDOWN_HTML_ANCHOR_DEFINITION_REGEX.sub(
            functools.partial(found_href, url_group_index=1),
            paragraph,
        )

    return transform_p_by_p_skipping_codeblocks(
        markdown,
        transform,
    )


def interpret_escapes(value: str) -> str:
    """Interpret Python literal escapes in a string.

    Replaces any standard escape sequences in value with their usual
    meanings as in ordinary Python string literals.
    """
    return value.encode('latin-1', 'backslashreplace').decode('unicode_escape')


def filter_inclusions(  # noqa: PLR0912
        start: str | None,
        end: str | None,
        text_to_include: str,
) -> tuple[str, bool, bool]:
    """Filter inclusions in a text.

    Manages inclusions from files using ``start`` and ``end`` directive
    arguments.
    """
    expected_start_not_found, expected_end_not_found = (False, False)
    new_text_to_include = ''

    if start is not None and end is None:
        start = interpret_escapes(start)
        if start not in text_to_include:
            expected_start_not_found = True
        else:
            new_text_to_include = text_to_include.split(
                start,
                maxsplit=1,
            )[1]
    elif start is None and end is not None:
        end = interpret_escapes(end)
        if end not in text_to_include:
            expected_end_not_found = True
            new_text_to_include = text_to_include
        else:
            new_text_to_include = text_to_include.split(
                end,
                maxsplit=1,
            )[0]
    elif start is not None and end is not None:
        start, end = interpret_escapes(start), interpret_escapes(end)
        if start not in text_to_include:
            expected_start_not_found = True
        if end not in text_to_include:
            expected_end_not_found = True

        start_split = text_to_include.split(start)
        text_parts = (
            start_split[1:]
            if len(start_split) > 1 else [text_to_include]
        )

        for start_text in text_parts:
            for i, end_text in enumerate(start_text.split(end)):
                if not i % 2:
                    new_text_to_include += end_text
    else:  # pragma: no cover
        new_text_to_include = text_to_include

    return (
        new_text_to_include,
        expected_start_not_found,
        expected_end_not_found,
    )


def _transform_negative_offset_func_factory(
        offset: int,
) -> Callable[[str], str]:
    abs_offset = abs(offset)

    def transform(line: str) -> str:
        try:
            if line[0] != '#':
                return line
        except IndexError:  # pragma: no cover
            # Note for pragma: all lines include a newline
            # so this exception is never raised in tests.
            return line
        stripped_line = line.lstrip('#')
        new_n_headings = max(len(line) - len(stripped_line) - abs_offset, 1)
        return '#' * new_n_headings + stripped_line

    return transform


def _transform_positive_offset_func_factory(
        offset: int,
) -> Callable[[str], str]:
    heading_prefix = '#' * offset

    def transform(line: str) -> str:
        try:
            if line[0] != '#':
                return line
        except IndexError:  # pragma: no cover
            return line
        return heading_prefix + line

    return transform


def increase_headings_offset(markdown: str, offset: int = 0) -> str:
    """Increases the headings depth of a snippet of Makdown content."""
    if not offset:  # pragma: no cover
        return markdown
    return transform_line_by_line_skipping_codeblocks(
        markdown,
        _transform_positive_offset_func_factory(offset) if offset > 0
        else _transform_negative_offset_func_factory(offset),
    )


def rstrip_trailing_newlines(content: str) -> str:
    """Removes trailing newlines from a string."""
    while content.endswith(('\n', '\r')):
        content = content.rstrip('\r\n')
    return content


def filter_paths(
        filepaths: Iterator[str] | list[str],
        ignore_paths: list[str],
) -> list[str]:
    """Filters a list of paths removing those defined in other list of paths.

    The paths to filter can be defined in the list of paths to ignore in
    several forms:

    - The same string.
    - Only the file name.
    - Only their direct directory name.
    - Their direct directory full path.

    Args:
        filepaths (list): Set of source paths to filter.
        ignore_paths (list): Paths that are ignored.

    Returns:
        list: Non filtered paths ordered alphabetically.
    """
    result = []
    for filepath in filepaths:
        # ignore by filepath
        if filepath in ignore_paths:
            continue

        # ignore by dirpath (relative or absolute)
        fp_split = filepath.split(os.sep)
        fp_split.pop()
        if (os.sep).join(fp_split) in ignore_paths:
            continue

        # ignore if is a directory
        try:
            if not stat.S_ISDIR(os.stat(filepath).st_mode):
                result.append(filepath)
        except (FileNotFoundError, OSError):  # pragma: no cover
            continue
    return result


def natural_sort_key(s: str) -> list[Any]:
    """Key function for natural sorting of strings."""
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split(r'(\d+)', s)]


def sort_paths(paths: list[str], order: OrderOption) -> list[str]:
    """Sort a list of paths in-place according to an order option."""
    ascending, order_type, order_by = order

    if order_type == 'random':
        import random  # noqa: PLC0415

        random.shuffle(paths)
        return paths

    key = None
    if order_type == 'alpha':
        if order_by == 'name':
            def key(p: str) -> str:
                return os.path.basename(p)
        elif order_by == 'extension':
            def key(p: str) -> str:
                return os.path.splitext(p)[1]
    elif order_type == 'natural':
        if order_by == 'extension':
            def key(p: str) -> str:
                return natural_sort_key(os.path.splitext(p)[1])  # type: ignore
        elif order_by == 'name':
            def key(p: str) -> str:
                return natural_sort_key(os.path.basename(p))  # type: ignore
        else:
            key = natural_sort_key  # type: ignore
    elif order_type == 'size':
        def key(p: str) -> int:  # type: ignore
            return os.path.getsize(p)
        ascending = not ascending  # larger files first
    elif order_type == 'mtime':
        def key(p: str) -> float:  # type: ignore
            return os.path.getmtime(p)
    elif order_type == 'ctime':
        def key(p: str) -> float:  # type: ignore
            return os.path.getctime(p)
    elif order_type == 'atime':
        def key(p: str) -> float:  # type: ignore
            return os.path.getatime(p)
    paths.sort(key=key, reverse=ascending)
    return paths


def _is_valid_url_scheme_char(c: str) -> bool:
    """Determine is a character is a valid URL scheme character.

    Valid characters are:

    ```
    abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-.
    ```
    """
    codepoint = ord(c)
    A = 65
    Z = 90
    a = 97
    z = 122
    zero = 48
    nine = 57
    dot = 46
    plus = 43
    minus = 45
    return (
        A <= codepoint <= Z
        or a <= codepoint <= z
        or zero <= codepoint <= nine
        or codepoint in (plus, minus, dot)
    )


def is_url(string: str) -> bool:
    """Determine if a string is an URL.

    The implementation has been adapted from `urllib.urlparse`.
    """
    i = string.find(':')
    if i <= 1:  # noqa: PLR2004 -> exclude C: or D: on Windows
        return False

    try:
        return all(_is_valid_url_scheme_char(string[j]) for j in range(i))
    except (IndexError, ValueError):  # pragma: no cover
        return False


def is_relative_path(string: str) -> bool:
    """Check if a string looks like a relative path."""
    try:
        return (
            string[0] == '.'
            and (
                string[1] == '/'
                or (string[1] == '.' and string[2] == '/')
            )
        )
    except IndexError:  # pragma: no cover
        return False


def is_absolute_path(string: str) -> bool:
    """Check if a string looks like an absolute path."""
    try:
        return string[0] == '/' or string[0] == os.sep
    except IndexError:  # pragma: no cover
        return False


def is_anchor(string: str) -> bool:
    """Check if a string looks like an anchor.

    An anchor is a string that starts with `#` and is not a relative path.
    """
    try:
        return string[0] == '#'
    except IndexError:  # pragma: no cover
        return False


def read_file(file_path: str, encoding: str) -> str:
    """Read a file and return its content."""
    f = open(file_path, encoding=encoding)  # noqa: SIM115
    content = f.read()
    f.close()
    return content


def read_url(
        url: str,
        http_cache: Cache | None,
        encoding: str = 'utf-8',
) -> Any:
    """Read an HTTP location and return its content."""
    from urllib.request import Request, urlopen  # noqa: PLC0415

    if http_cache is not None:
        cached_content = http_cache.get_(url, encoding)
        if cached_content is not None:
            return cached_content
    with urlopen(Request(url)) as response:
        content = response.read().decode(encoding)
    if http_cache is not None:
        http_cache.set_(url, content, encoding)
    return content


def safe_os_path_relpath(path: str, start: str) -> str:
    """Return the relative path of a file from a start directory.

    Safe version of `os.path.relpath` that catches possible `ValueError`
    exceptions and returns the original path in case of error.
    On Windows, `ValueError` is raised when `path` and `start` are on
    different drives.
    """
    try:
        return os.path.relpath(path, start)
    except ValueError:  # pragma: no cover
        return path


def file_lineno_message(
        page_src_path: str | None,
        docs_dir: str,
        lineno: int,
) -> str:
    """Return a message with the file path and line number."""
    if page_src_path is None:  # pragma: no cover
        return f'generated page content (line {lineno})'
    return (
        f'{safe_os_path_relpath(page_src_path, docs_dir)}'
        f':{lineno}'
    )


def lineno_from_content_start(content: str, start: int) -> int:
    """Return the line number of the first line of ``start`` in ``content``."""
    return content[:start].count('\n') + 1



"""``include`` directive tests."""

import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown


@pytest.mark.parametrize(
    (
        'includer_schema',
        'content_to_include',
        'expected_result',
        'expected_warnings_schemas',
    ),
    (
        pytest.param(
            '# Header\n\n{% include "{filepath}" %}\n',
            'This must be included.',
            '# Header\n\nThis must be included.\n',
            [],
            id='simple-case',
        ),

        # Newline at the end of the included content
        pytest.param(
            '# Header\n\n{% include "{filepath}" %}\n',
            'This must be included.\n',
            '# Header\n\nThis must be included.\n\n',
            [],
            id='newline-at-end-of-included',
        ),

        # Start and end options
        pytest.param(
            '''# Header

{%
  include "{filepath}"
  start="start here"
  end="end here"
%}
''',
            '''This must be ignored.
start hereThis must be included.end here
This must be ignored also.
''',
            '''# Header

This must be included.
''',
            [],
            id='start/end',
        ),

        # Start and end options with escaped special characters
        pytest.param(
            '''# Header

{%
  include "{filepath}"
  start="\\tstart here"
  end="\\tend here"
%}
''',
            '''This must be ignored.
\tstart hereThis must be included.\tend here
This must be ignored also.
''',
            '''# Header

This must be included.
''',
            [],
            id='start/end (escaped special characters)',
        ),

        # Start and end options with unescaped special characters
        pytest.param(
            '''# Header

{%
  include "{filepath}"
  start="\tstart here"
  end="\tend here"
%}
''',
            '''This must be ignored.
\tstart hereThis must be included.\tend here
This must be ignored also.
''',
            '''# Header

This must be included.
''',
            [],
            id='start/end (unescaped special characters)',
        ),

        # Multiples start and end matchs
        pytest.param(
            '''{%
  include-markdown "{filepath}"
  start="<!--start-tag-->"
  end="<!--end-tag-->"
%}''',
            '''Some text

<!--start-tag-->
This should be included.
<!--end-tag-->

This shouldn't be included.

<!--start-tag-->
This should be included also.
<!--end-tag-->

Here some text
that should be ignored.

<!--start-->
<!--end-->

Etc
<!--start-tag-->
This should be included even if hasn't defined after end tag.
''',
            '''
This should be included.

This should be included also.

This should be included even if hasn't defined after end tag.
''',
            [],
            id='multiple-start-end-matchs',
        ),

        # Don't specify end and finds start in included content
        pytest.param(
            '''Foo
{%
    include "{filepath}"
    start="<!--start-->"
%}''',
            '''Some text

<!--start-->
More text
''',
            '''Foo

More text
''',
            [],
            id='start=foo-end=None',
        ),

        # Don't specify start and finds end in included content
        pytest.param(
            '''Foo
{%
    include "{filepath}"
    end="<!--end-->"
%}''',
            '''
Some text
<!--end-->
More text
''',
            '''Foo

Some text
''',
            [],
            id='start=None-end=foo',
        ),

        # Don't specify end but not finds start in included content
        pytest.param(
            '''Foo

{%
    include "{filepath}"
    start="<!--start-->"
%}''',
            '''Some text
''',
            '''Foo

''',
            [
                (
                    "Delimiter start '<!--start-->' of 'include'"
                    ' directive at {filepath}:3'
                    ' not detected in the file {included_file}'
                ),
            ],
            id='start=foo (not found)-end=None',
        ),

        # Don't specify start but not finds end in included content
        pytest.param(
            '''Foo
{%
    include "{filepath}"
    end="<!--end-->"
%}''',
            '''
Some text
''',
            '''Foo

Some text
''',
            [
                (
                    "Delimiter end '<!--end-->' of 'include'"
                    ' directive at {filepath}:2'
                    ' not detected in the file {included_file}'
                ),
            ],
            id='start=None-end=foo (not found)',
        ),

        # Preserve included indent
        pytest.param(
            '''1. Ordered list item
    {%
      include "{filepath}"
      preserve-includer-indent=false
    %}''',
            '''- Unordered sublist item
    - Other unordered sublist item''',
            '''1. Ordered list item
    - Unordered sublist item
    - Other unordered sublist item''',
            [],
            id='preserve included indent',
        ),

        # Preserve includer indent
        pytest.param(
            '''1. Ordered list item
    {%
      include "{filepath}"
    %}''',
            '''- First unordered sublist item
- Second unordered sublist item
- Third unordered sublist item''',
            '''1. Ordered list item
    - First unordered sublist item
    - Second unordered sublist item
    - Third unordered sublist item''',
            [],
            id='preserve includer indent',
        ),

        # Custom options ordering
        pytest.param(
            '''1. Ordered list item
    {%
      include "{filepath}"
      preserve-includer-indent=true
      end="<!--end-->"
      start="<!--start-->"
    %}''',
            '''<!--start-->- First unordered sublist item
- Second unordered sublist item<!--end-->
- Third unordered sublist item''',
            '''1. Ordered list item
    - First unordered sublist item
    - Second unordered sublist item''',
            [],
            id='custom options ordering',
        ),

        # Content unindentation
        pytest.param(
            '''# Header

{%
  include "{filepath}"
  dedent=true
%}
''',
            '''    - Foo
    - Bar
        - Baz''',
            '''# Header

- Foo
- Bar
    - Baz
''',
            [],
            id='dedent=true',
        ),

        # Include from URL
        pytest.param(
            '''# Header

{% include "https://raw.githubusercontent.com/mondeja/mkdocs-include-markdown-plugin/master/examples/basic/docs/included.md" %}
''',  # noqa: E501
            '(not used)\n',
            '''# Header

Some ignored content.

<--start-->

Some included content.

''',
            [],
            id='url',
        ),

        # Content unindentation + preserve includer indent
        pytest.param(
            '''# Header

    {%
      include "{filepath}"
      dedent=true
      preserve-includer-indent=true
    %}
''',
            '''        - Foo
        - Bar
            - Baz''',
            '''# Header

    - Foo
    - Bar
        - Baz
''',
            [],
            id='dedent=true,preserve-includer-indent=true',
        ),
        pytest.param(
            'Foo{% include "{filepath}" trailing-newlines=false %}',
            '\n',
            'Foo',
            [],
            id='rstrip-unix-trailing-newlines',
        ),
        pytest.param(
            'Foo{% include "{filepath}" trailing-newlines=false %}',
            '\r\n\r\n',
            'Foo',
            [],
            id='rstrip-windows-trailing-newlines',
        ),
    ),
)
def test_include(
    includer_schema,
    content_to_include,
    expected_result,
    expected_warnings_schemas,
    page,
    plugin,
    caplog,
    tmp_path,
):
    included_file = tmp_path / 'included.md'
    includer_file = tmp_path / 'includer.md'

    included_file.write_text(content_to_include)
    includer_file.write_text(
        content_to_include.replace('{filepath}', included_file.as_posix()),
    )

    # assert content
    page_content = includer_schema.replace(
        '{filepath}', included_file.as_posix(),
    )
    includer_file.write_text(page_content)

    assert on_page_markdown(
        page_content, page(includer_file), tmp_path, plugin,
    ) == expected_result

    # assert warnings
    expected_warnings_schemas = expected_warnings_schemas or []
    expected_warnings = [
        msg_schema.replace(
            '{filepath}',
            str(includer_file.relative_to(tmp_path)),
        ).replace(
            '{included_file}',
            str(included_file.relative_to(tmp_path)),
        ) for msg_schema in expected_warnings_schemas
    ]

    for record in caplog.records:
        assert record.msg in expected_warnings
    assert len(expected_warnings_schemas) == len(caplog.records)



import pytest
from mkdocs.exceptions import PluginError

from mkdocs_include_markdown_plugin.directive import get_order_option_regex
from mkdocs_include_markdown_plugin.plugin import IncludeMarkdownPlugin
from testing_helpers import FakeConfig


def test_invalid_order_setting():
    plugin = IncludeMarkdownPlugin()
    plugin.config = FakeConfig(order='invalid-order')
    with pytest.raises(PluginError) as exc:
        plugin.on_config({})
    regex = get_order_option_regex()
    assert (
        "Invalid value 'invalid-order' for the 'order' global setting."
        f" Order must be a string that matches the regex '{regex.pattern}'."
    ) in str(exc.value)


def test_valid_order_setting():
    plugin = IncludeMarkdownPlugin()
    plugin.config = FakeConfig(order='alpha-name')
    assert plugin.on_config({}) is not None



"""Tests for multiple inclusions across directives."""

import os

import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives, unix_only


@unix_only
@parametrize_directives
def test_glob_include_absolute(directive, page, tmp_path, plugin):
    includer_file = tmp_path / 'includer.txt'
    included_01_file = tmp_path / 'included_01.txt'
    included_02_file = tmp_path / 'included_02.txt'

    includer_file_content = f'''foo

{{%
  {directive} "./included*.txt"
%}}

<!-- with absolute path -->
{{%
  {directive} "{tmp_path}{os.sep}included*.txt"
%}}
'''

    included_01_content = 'bar'
    included_02_content = 'baz'

    includer_file.write_text(includer_file_content)
    included_01_file.write_text(included_01_content)
    included_02_file.write_text(included_02_content)

    expected_result = '''foo

barbaz

<!-- with absolute path -->
barbaz
'''

    assert on_page_markdown(
        includer_file_content, page(includer_file), tmp_path, plugin,
    ) == expected_result


@unix_only
@parametrize_directives
def test_glob_include_fallback(directive, page, tmp_path, plugin):
    includer_file = tmp_path / 'includer.txt'
    includes_dir = tmp_path / 'includes'
    includes_dir.mkdir()
    included_01_file = includes_dir / 'included_01.txt'
    included_02_file = includes_dir / 'included_02.txt'

    includer_file_content = f'''foo

{{%
  {directive} "includes/*.txt"
%}}
'''

    included_01_content = 'bar'
    included_02_content = 'baz'

    includer_file.write_text(includer_file_content)
    included_01_file.write_text(included_01_content)
    included_02_file.write_text(included_02_content)

    expected_result = '''foo

barbaz
'''

    assert on_page_markdown(
        includer_file_content, page(includer_file), tmp_path, plugin,
    ) == expected_result


@parametrize_directives
@pytest.mark.parametrize(
    (
        'includer_content',
        'expected_warnings_schemas',
    ),
    (
        pytest.param(
            '''{%
  {directive} "./included*.txt"
  start="<!-- start-2 -->"
  end="<!-- end-2 -->"
%}

{%
  {directive} "./included*.txt"
  start="<!-- start-1 -->"
  end="<!-- end-1 -->"
%}
''',
            [],
            id='start-end',
        ),
        pytest.param(
            '''{%
  {directive} "./included*.txt"
  end="<!-- end-2 -->"
%}
''',
            [],
            id='end',
        ),

        # both start and end specified but not found in files to include
        pytest.param(
            '''{%
  {directive} "./included*.txt"
  start="<!-- start-not-found-2 -->"
  end="<!-- end-not-found-2 -->"
%}

{%
  {directive} "./included*.txt"
  start="<!-- start-not-found-1 -->"
  end="<!-- end-not-found-1 -->"
%}
''',
            [
                (
                    "Delimiter end '<!-- end-not-found-1 -->'"
                    " of '{directive}' directive"
                    ' at {includer_file}:9 not detected in'
                    ' the files {included_file_01}, {included_file_02}'
                ),
                (
                    "Delimiter end '<!-- end-not-found-2 -->'"
                    " of '{directive}' directive"
                    ' at {includer_file}:3 not detected in'
                    ' the files {included_file_01}, {included_file_02}'
                ),
                (
                    "Delimiter start '<!-- start-not-found-1 -->'"
                    " of '{directive}' directive"
                    ' at {includer_file}:9 not detected in'
                    ' the files {included_file_01}, {included_file_02}'
                ),
                (
                    "Delimiter start '<!-- start-not-found-2 -->'"
                    " of '{directive}' directive"
                    ' at {includer_file}:3 not detected in'
                    ' the files {included_file_01}, {included_file_02}'
                ),
            ],
            id='start-end-not-found',
        ),
    ),
)
def test_glob_include(
    includer_content,
    directive,
    expected_warnings_schemas,
    page,
    plugin,
    caplog,
    tmp_path,
):
    includer_file = tmp_path / 'includer.txt'
    included_01_file = tmp_path / 'included_01.txt'
    included_02_file = tmp_path / 'included_02.txt'

    includer_file_content = f'''foo

{includer_content.replace('{directive}', directive)}
'''

    included_01_content = '''This 01 must appear only without specifying start.
<!-- start-1 -->
bar
<!-- end-1 -->
This 01 must appear only without specifying end.
'''
    included_02_content = '''This 02 must appear only without specifying start.
<!-- start-2 -->
baz
<!-- end-2 -->
This 02 must appear only without specifying end.
'''

    includer_file.write_text(includer_file_content)
    included_01_file.write_text(included_01_content)
    included_02_file.write_text(included_02_content)

    on_page_markdown(
        includer_file_content, page(includer_file), tmp_path, plugin,
    )

    # assert warnings
    expected_warnings_schemas = expected_warnings_schemas or []
    expected_warnings = [
        msg_schema.replace(
            '{includer_file}',
            str(includer_file.relative_to(tmp_path)),
        ).replace(
            '{included_file_01}',
            str(included_01_file.relative_to(tmp_path)),
        ).replace(
            '{included_file_02}',
            str(included_02_file.relative_to(tmp_path)),
        ).replace('{directive}', directive)
        for msg_schema in expected_warnings_schemas
    ]

    for record in caplog.records:
        assert record.msg in expected_warnings
    assert len(expected_warnings_schemas) == len(caplog.records)



"""Plugin entry point."""

from __future__ import annotations

from collections.abc import Callable
from functools import cached_property
from typing import TYPE_CHECKING

from mkdocs.exceptions import PluginError
from mkdocs.plugins import BasePlugin, event_priority


if TYPE_CHECKING:  # pragma: no cover
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.livereload import LiveReloadServer
    from mkdocs.structure.files import Files
    from mkdocs.structure.pages import Page

from mkdocs_include_markdown_plugin.cache import Cache, initialize_cache
from mkdocs_include_markdown_plugin.config import PluginConfig
from mkdocs_include_markdown_plugin.directive import (
    get_order_option_regex,
)
from mkdocs_include_markdown_plugin.event import (
    on_page_markdown as _on_page_markdown,
)
from mkdocs_include_markdown_plugin.files_watcher import FilesWatcher


class IncludeMarkdownPlugin(BasePlugin[PluginConfig]):
    _cache: Cache | None = None
    _server: LiveReloadServer | None = None

    def on_config(self, config: MkDocsConfig) -> MkDocsConfig:
        if self.config.cache > 0:
            cache = initialize_cache(self.config.cache, self.config.cache_dir)
            if cache is None:
                raise PluginError(
                    'Either `cache_dir` global setting must be configured or'
                    ' `platformdirs` package is required to use the'
                    ' `cache` option. Install mkdocs-include-markdown-plugin'
                    " with the 'cache' extra to install `platformdirs`.",
                )
            self._cache = cache

        if '__default' not in self.config.directives:  # pragma: no cover
            for directive in self.config.directives:
                if directive not in ('include', 'include-markdown'):
                    raise PluginError(
                        f"Invalid directive name '{directive}' at 'directives'"
                        ' global setting. Valid values are "include" and'
                        ' "include-markdown".',
                    )

        if self.config.order != 'alpha-path':
            regex = get_order_option_regex()
            if not regex.match(self.config.order):
                raise PluginError(
                    f"Invalid value '{self.config.order}' for the 'order'"
                    ' global setting. Order must be a string'
                    f" that matches the regex '{regex.pattern}'.",
                )

        return config

    @cached_property
    def _files_watcher(self) -> FilesWatcher:
        return FilesWatcher()

    def _update_watched_files(self) -> None:  # pragma: no cover
        """Function executed on server reload.

        At this execution point, the ``self._server`` attribute must be set.
        """
        watcher, server = self._files_watcher, self._server

        # unwatch previous watched files not needed anymore
        for file_path in watcher.prev_included_files:
            if file_path not in watcher.included_files:
                server.unwatch(file_path)  # type: ignore
        watcher.prev_included_files = watcher.included_files[:]

        # watch new included files
        for file_path in watcher.included_files:
            server.watch(file_path, recursive=False)  # type: ignore
        watcher.included_files = []

    def on_page_content(
            self,
            html: str,
            page: Page,  # noqa: ARG002
            config: MkDocsConfig,  # noqa: ARG002
            files: Files,  # noqa: ARG002
    ) -> str:
        if self._server is not None:  # pragma: no cover
            self._update_watched_files()
        return html

    def on_serve(
            self,
            server: LiveReloadServer,
            config: MkDocsConfig,  # noqa: ARG002
            builder: Callable,  # noqa: ARG002
    ) -> None:
        if self._server is None:  # pragma: no cover
            self._server = server
            self._update_watched_files()

    @event_priority(100)
    def on_page_markdown(
            self,
            markdown: str,
            page: Page,
            config: MkDocsConfig,
            files: Files,  # noqa: ARG002
    ) -> str:
        return _on_page_markdown(
            markdown,
            page,
            config.docs_dir,
            plugin=self,
        )



"""Installation using setup.py is no longer supported.
Use `pip install .` instead.
"""

import sys

from setuptools import setup


sys.exit(__doc__)

# Fake reference so GitHub still considers it a real package
# for statistics purposes:
setup(
    name='mkdocs-include-markdown-plugin',
)



import pytest

from mkdocs_include_markdown_plugin.event import on_page_markdown
from testing_helpers import parametrize_directives, unix_only


@parametrize_directives
def test_encoding(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'included.md'
    page_to_include_file.write_text('''Á
<!-- start -->
Content to include
<!-- end -->
É
''')

    with pytest.raises(UnicodeDecodeError):
        on_page_markdown(
            f'''{{%
    {directive} "{page_to_include_file}"
    start='<!-- start -->'
    end="<!-- end -->"
    encoding="ascii"
%}}''',
            page(tmp_path / 'includer.md'),
            tmp_path,
            plugin,
        )


@unix_only
@parametrize_directives
def test_default_encoding(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'included.md'
    page_to_include_file.write_text('''Á
<!-- start -->
Content to include
<!-- end -->
É
''')

    result = on_page_markdown(
        f'''{{%
  {directive} "{page_to_include_file}"
  start="<!-- start -->"
  end="<!-- end -->"
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    )
    assert result == '\nContent to include\n'


@unix_only
@parametrize_directives
def test_explicit_default_encoding(directive, page, tmp_path, plugin):
    page_to_include_file = tmp_path / 'included.md'
    page_to_include_file.write_text('''Á
<!-- start -->
Content to include
<!-- end -->
É
''')

    result = on_page_markdown(
        f'''{{%
  {directive} "{page_to_include_file}"
  start='<!-- start -->'
  end="<!-- end -->"
  encoding="utf-8"
%}}''',
        page(tmp_path / 'includer.md'),
        tmp_path,
        plugin,
    )
    assert result == '\nContent to include\n'
